@@ -1,25 +1,31 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines utilities for internal use. This is not a public interface
 and may be changed without notice.
 """
 
+from __future__ import annotations
+
 import datetime
 import glob
 import itertools
 import numbers
 import os
-import platform
 import re
 import warnings
 from collections.abc import Iterable
-from urllib.parse import unquote, urlparse
+from pathlib import Path
+from typing import IO
+from urllib.parse import urlparse
 
+import fsspec
 import numpy
-import packaging
+import packaging.version
 
-win = platform.system().lower().startswith("win")
+import uproot.source.chunk
+import uproot.source.fsspec
+import uproot.source.object
 
 
 def tobytes(array):
@@ -33,7 +39,7 @@ def tobytes(array):
         return array.tostring()
 
 
-def isint(x):
+def isint(x) -> bool:
     """
     Returns True if and only if ``x`` is an integer (including NumPy, not
     including bool).
@@ -43,7 +49,7 @@ def isint(x):
     )
 
 
-def isnum(x):
+def isnum(x) -> bool:
     """
     Returns True if and only if ``x`` is a number (including NumPy, not
     including bool).
@@ -53,14 +59,7 @@ def isnum(x):
     )
 
 
-def isstr(x):
-    """
-    Returns True if and only if ``x`` is a string (including Python 2 unicode).
-    """
-    return isinstance(x, str)
-
-
-def ensure_str(x):
+def ensure_str(x) -> str:
     """
     Ensures that ``x`` is a string (decoding with 'surrogateescape' if necessary).
     """
@@ -77,29 +76,51 @@ def ensure_numpy(array, types=(numpy.bool_, numpy.integer, numpy.floating)):
     Returns an ``np.ndarray`` if ``array`` can be converted to an array of the
     desired type and raises TypeError if it cannot.
     """
+    import uproot
+
+    awkward = uproot.extras.awkward()
     with warnings.catch_warnings():
-        warnings.simplefilter("error", numpy.VisibleDeprecationWarning)
-        try:
-            out = numpy.asarray(array)
-        except (ValueError, numpy.VisibleDeprecationWarning) as err:
-            raise TypeError("cannot be converted to a NumPy array") from err
+        warnings.simplefilter(
+            "error", getattr(numpy, "exceptions", numpy).VisibleDeprecationWarning
+        )
+        if isinstance(array, awkward.contents.Content):
+            out = awkward.to_numpy(array)
+        else:
+            try:
+                out = numpy.asarray(array)
+            except (ValueError, numpy.VisibleDeprecationWarning) as err:
+                raise TypeError("cannot be converted to a NumPy array") from err
         if not issubclass(out.dtype.type, types):
             raise TypeError(f"cannot be converted to a NumPy array of type {types}")
         return out
 
 
-def parse_version(version):
+def is_file_like(
+    obj, readable: bool = False, writable: bool = False, seekable: bool = False
+) -> bool:
+    return (
+        all(
+            callable(getattr(obj, attr, None))
+            for attr in ("read", "write", "seek", "tell", "flush")
+        )
+        and (not readable or not hasattr(obj, "readable") or obj.readable())
+        and (not writable or not hasattr(obj, "writable") or obj.writable())
+        and (not seekable or not hasattr(obj, "seekable") or obj.seekable())
+    )
+
+
+def parse_version(version: str):
     """
     Converts a semver string into a Version object that can be compared with
     ``<``, ``>=``, etc.
 
-    Currently implemented using ``packaging.version.parse``
+    Currently implemented using ``packaging.Version``
     (exposing that library in the return type).
     """
     return packaging.version.parse(version)
 
 
-def from_module(obj, module_name):
+def from_module(obj, module_name: str) -> bool:
     """
     Returns True if ``obj`` is an instance of a class from a module
     given by name.
@@ -138,7 +159,7 @@ def _regularize_filter_regex_flags(flags):
     return flagsbyte
 
 
-def no_filter(x):
+def no_filter(x) -> bool:
     """
     A filter that accepts anything (always returns True).
     """
@@ -157,7 +178,7 @@ def regularize_filter(filter):
         return no_filter
     elif callable(filter):
         return filter
-    elif isstr(filter):
+    elif isinstance(filter, str):
         m = _regularize_filter_regex.match(filter)
         if m is not None:
             regex, flags = m.groups()
@@ -173,7 +194,7 @@ def regularize_filter(filter):
     else:
         raise TypeError(
             "filter must be None, callable, a regex string between slashes, or a "
-            "glob pattern, not {}".format(repr(filter))
+            f"glob pattern, not {filter!r}"
         )
 
 
@@ -198,7 +219,7 @@ def regularize_rename(rename):
     elif callable(rename):
         return rename
 
-    elif isstr(rename):
+    elif isinstance(rename, str):
         m = _regularize_filter_regex_rename.match(rename)
         if m is not None:
             regex, trans, flags = m.groups()
@@ -213,7 +234,7 @@ def regularize_rename(rename):
     elif isinstance(rename, Iterable) and not isinstance(rename, bytes):
         rules = []
         for x in rename:
-            if isstr(x):
+            if isinstance(x, str):
                 m = _regularize_filter_regex_rename.match(x)
                 if m is not None:
                     regex, trans, flags = m.groups()
@@ -234,7 +255,7 @@ def regularize_rename(rename):
             def applyrules(x):
                 for matcher, trans in rules:
                     if matcher.search(x) is not None:
-                        return matcher.sub(trans, x)  # noqa: B023
+                        return matcher.sub(trans, x)
                 else:
                     return x
 
@@ -242,148 +263,104 @@ def regularize_rename(rename):
 
     raise TypeError(
         "rename must be None, callable, a '/from/to/' regex, an iterable of "
-        "regex rules, or a dict, not {}".format(repr(rename))
+        f"regex rules, or a dict, not {rename!r}"
     )
 
 
+_fix_url_path = re.compile(r"^((file|https?|root):/)([^/])", re.I)
+
+
 def regularize_path(path):
     """
     Converts pathlib Paths into plain string paths (for all versions of Python).
     """
     if isinstance(path, getattr(os, "PathLike", ())):
-        path = os.fspath(path)
+        path = _fix_url_path.sub(r"\1/\3", os.fspath(path))
 
     elif hasattr(path, "__fspath__"):
-        path = path.__fspath__()
+        path = _fix_url_path.sub(r"\1/\3", path.__fspath__())
 
     elif path.__class__.__module__ == "pathlib":
         import pathlib
 
         if isinstance(path, pathlib.Path):
-            path = str(path)
+            path = _fix_url_path.sub(r"\1/\3", str(path))
 
     return path
 
 
-_windows_drive_letter_ending = re.compile(r".*\b[A-Za-z]$")
-_windows_absolute_path_pattern = re.compile(r"^[A-Za-z]:[\\/]")
-_windows_absolute_path_pattern_slash = re.compile(r"^[\\/][A-Za-z]:[\\/]")
-_might_be_port = re.compile(r"^[0-9].*")
-
-
-def file_object_path_split(path):
+def file_object_path_split(urlpath: str) -> tuple[str, str | None]:
     """
     Split a path with a colon into a file path and an object-in-file path.
+
+    Args:
+        urlpath: The path to split. Example: ``"https://localhost:8888/file.root:tree"``
+
+    Returns:
+        A tuple of the file path and the object-in-file path. If there is no
+        object-in-file path, the second element is ``None``.
+        Example: ``("https://localhost:8888/file.root", "tree")``
     """
-    path = regularize_path(path)
 
-    try:
-        index = path.rindex(":")
-    except ValueError:
-        return path, None
-    else:
-        file_path, object_path = path[:index], path[index + 1 :]
+    urlpath: str = regularize_path(urlpath).strip()
+    obj = None
 
-        if (
-            _might_be_port.match(object_path) is not None
-            and urlparse(file_path).path == ""
-        ):
-            return path, None
+    separator = "::"
+    parts = urlpath.split(separator)
+    object_regex = re.compile(r"(.+\.root):(.*$)")
+    for i, part in enumerate(reversed(parts)):
+        match = object_regex.match(part)
+        if match:
+            obj = re.sub(r"/+", "/", match.group(2).strip().lstrip("/")).rstrip("/")
+            parts[-i - 1] = match.group(1)
+            break
 
-        file_path = file_path.rstrip()
-        object_path = object_path.lstrip()
-
-        if file_path.upper() in ("FILE", "HTTP", "HTTPS", "ROOT"):
-            return path, None
-        elif win and _windows_drive_letter_ending.match(file_path) is not None:
-            return path, None
-        else:
-            return file_path, object_path
+    urlpath = separator.join(parts)
+    return urlpath, obj
 
 
-_remote_schemes = ["ROOT", "HTTP", "HTTPS"]
-_schemes = ["FILE"] + _remote_schemes
-
-
-def file_path_to_source_class(file_path, options):
+def file_path_to_source_class(
+    file_path_or_object: str | Path | IO, options: dict
+) -> tuple[type[uproot.source.chunk.Source], str | IO]:
     """
     Use a file path to get the :doc:`uproot.source.chunk.Source` class that would read it.
+
+    Returns a tuple of (class, file_path) where the class is a subclass of :doc:`uproot.source.chunk.Source`.
     """
-    import uproot.source.chunk
 
-    file_path = regularize_path(file_path)
+    file_path_or_object: str | IO = regularize_path(file_path_or_object)
 
-    if (
-        not isstr(file_path)
-        and hasattr(file_path, "read")
-        and hasattr(file_path, "seek")
+    source_cls = options["handler"]
+    if source_cls is not None and not (
+        isinstance(source_cls, type)
+        and issubclass(source_cls, uproot.source.chunk.Source)
     ):
-        out = options["object_handler"]
-        if not (isinstance(out, type) and issubclass(out, uproot.source.chunk.Source)):
-            raise TypeError(
-                "'object_handler' is not a class object inheriting from Source: "
-                + repr(out)
-            )
-        return out, file_path
+        raise TypeError(
+            f"'handler' is not a class object inheriting from Source: {source_cls!r}"
+        )
 
-    windows_absolute_path = None
-    if win:
-        if _windows_absolute_path_pattern.match(file_path) is not None:
-            windows_absolute_path = file_path
-
-    parsed_url = urlparse(file_path)
-    if parsed_url.scheme.upper() == "FILE":
-        parsed_url_path = unquote(parsed_url.path)
-    else:
-        parsed_url_path = parsed_url.path
-
-    if win and windows_absolute_path is None:
-        if _windows_absolute_path_pattern.match(parsed_url_path) is not None:
-            windows_absolute_path = parsed_url_path
-        elif _windows_absolute_path_pattern_slash.match(parsed_url_path) is not None:
-            windows_absolute_path = parsed_url_path[1:]
-
-    if (
-        parsed_url.scheme.upper() == "FILE"
-        or len(parsed_url.scheme) == 0
-        or windows_absolute_path is not None
+    # Infer the source class from the file path
+    if all(
+        callable(getattr(file_path_or_object, attr, None)) for attr in ("read", "seek")
     ):
-        if windows_absolute_path is None:
-            if parsed_url.netloc.upper() == "LOCALHOST":
-                file_path = parsed_url_path
-            else:
-                file_path = parsed_url.netloc + parsed_url_path
-        else:
-            file_path = windows_absolute_path
-
-        out = options["file_handler"]
-        if not (isinstance(out, type) and issubclass(out, uproot.source.chunk.Source)):
+        # need a very soft object check for ubuntu python3.8 pyroot ci tests, cannot use uproot._util.is_file_like
+        if (
+            source_cls is not None
+            and source_cls is not uproot.source.object.ObjectSource
+        ):
             raise TypeError(
-                "'file_handler' is not a class object inheriting from Source: "
-                + repr(out)
+                f"'handler' is not ObjectSource for a file-like object: {source_cls!r}"
             )
-        return out, os.path.expanduser(file_path)
-
-    elif parsed_url.scheme.upper() == "ROOT":
-        out = options["xrootd_handler"]
-        if not (isinstance(out, type) and issubclass(out, uproot.source.chunk.Source)):
-            raise TypeError(
-                "'xrootd_handler' is not a class object inheriting from Source: "
-                + repr(out)
-            )
-        return out, file_path
-
-    elif parsed_url.scheme.upper() == "HTTP" or parsed_url.scheme.upper() == "HTTPS":
-        out = options["http_handler"]
-        if not (isinstance(out, type) and issubclass(out, uproot.source.chunk.Source)):
-            raise TypeError(
-                "'http_handler' is not a class object inheriting from Source: "
-                + repr(out)
-            )
-        return out, file_path
-
+        return uproot.source.object.ObjectSource, file_path_or_object
+    elif isinstance(file_path_or_object, str):
+        source_cls = (
+            uproot.source.fsspec.FSSpecSource if source_cls is None else source_cls
+        )
+        return source_cls, file_path_or_object
     else:
-        raise ValueError(f"URI scheme not recognized: {file_path}")
+        raise TypeError(
+            f"file_path is not a string or file-like object: {file_path_or_object!r}"
+        )
 
 
 if isinstance(__builtins__, dict):
@@ -399,15 +376,12 @@ else:
 
 
 def _file_not_found(files, message=None):
-    if message is None:
-        message = ""
-    else:
-        message = " (" + message + ")"
+    message = "" if message is None else " (" + message + ")"
 
     return _FileNotFoundError(
-        """file not found{0}
+        f"""file not found{message}
 
-    {1}
+    {files!r}
 
 Files may be specified as:
    * str/bytes: relative or absolute filesystem path or URL, without any colons
@@ -426,18 +400,16 @@ Functions that accept many files (uproot.iterate, etc.) also allow:
          Example: {{"/data_v1/*.root": "ttree_v1", "/data_v2/*.root": "ttree_v2"}}
    * already-open TTree objects.
    * iterables of the above.
-""".format(
-            message, repr(files)
-        )
+"""
     )
 
 
-def memory_size(data, error_message=None):
+def memory_size(data, error_message=None) -> int:
     """
     Regularizes strings like '## kB' and plain integer number of bytes to
     an integer number of bytes.
     """
-    if isstr(data):
+    if isinstance(data, str):
         m = re.match(
             r"^\s*([+-]?(\d+(\.\d*)?|\.\d+)(e[+-]?\d+)?)\s*([kmgtpezy]?b)\s*$",
             data,
@@ -491,6 +463,36 @@ def memory_size(data, error_message=None):
         raise TypeError(error_message)
 
 
+def trim_final(basket_arrays, entry_start, entry_stop, entry_offsets, library, branch):
+    """
+    Trims the output from a basket_array function and outputs the un-concatenated list of elements.
+    """
+    trimmed = []
+    start = entry_offsets[0]
+    for basket_num, stop in enumerate(entry_offsets[1:]):
+        if start <= entry_start and entry_stop <= stop:
+            local_start = entry_start - start
+            local_stop = entry_stop - start
+            trimmed.append(basket_arrays[basket_num][local_start:local_stop])
+
+        elif start <= entry_start < stop:
+            local_start = entry_start - start
+            local_stop = stop - start
+            trimmed.append(basket_arrays[basket_num][local_start:local_stop])
+
+        elif start <= entry_stop <= stop:
+            local_start = 0
+            local_stop = entry_stop - start
+            trimmed.append(basket_arrays[basket_num][local_start:local_stop])
+
+        elif entry_start < stop and start <= entry_stop:
+            trimmed.append(basket_arrays[basket_num])
+
+        start = stop
+
+    return trimmed
+
+
 def new_class(name, bases, members):
     """
     Create a new class object with ``type(name, bases, members)`` and put it in
@@ -520,31 +522,35 @@ def awkward_form(model, file, context):
 
         if model not in _primitive_awkward_form:
             if model == numpy.dtype(numpy.bool_) or model == numpy.dtype(bool):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"bool"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"bool"')
             elif model == numpy.dtype(numpy.int8):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"int8"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"int8"')
             elif model == numpy.dtype(numpy.uint8):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"uint8"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"uint8"')
             elif model == numpy.dtype(numpy.int16):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"int16"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"int16"')
             elif model == numpy.dtype(numpy.uint16):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"uint16"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"uint16"')
             elif model == numpy.dtype(numpy.int32):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"int32"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"int32"')
             elif model == numpy.dtype(numpy.uint32):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"uint32"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"uint32"')
             elif model == numpy.dtype(numpy.int64):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"int64"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"int64"')
             elif model == numpy.dtype(numpy.uint64):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson('"uint64"')
+                _primitive_awkward_form[model] = awkward.forms.from_json('"uint64"')
             elif model == numpy.dtype(numpy.float32):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson(
-                    '"float32"'
-                )
+                _primitive_awkward_form[model] = awkward.forms.from_json('"float32"')
             elif model == numpy.dtype(numpy.float64):
-                _primitive_awkward_form[model] = awkward.forms.Form.fromjson(
-                    '"float64"'
-                )
+                _primitive_awkward_form[model] = awkward.forms.from_json('"float64"')
+            elif model.fields is not None:
+                fields = []
+                contents = []
+                for field, (dtype, _) in model.fields.items():
+                    fields.append(field)
+                    contents.append(awkward_form(dtype, file, context))
+                # directly return; don't cache RecordForms in _primitive_awkward_form
+                return awkward.forms.RecordForm(contents, fields)
             else:
                 raise AssertionError(f"{model!r}: {type(model)}")
 
@@ -554,116 +560,6 @@ def awkward_form(model, file, context):
         return model.awkward_form(file, context)
 
 
-def awkward_form_remove_uproot(awkward, form):
-    """
-    Remove the "uproot" parameters from an ``ak.forms.Form``.
-    """
-    parameters = dict(form.parameters)
-    parameters.pop("uproot", None)
-    if isinstance(form, awkward.forms.BitMaskedForm):
-        return awkward.forms.BitMaskedForm(
-            form.mask,
-            awkward_form_remove_uproot(awkward, form.content),
-            form.valid_when,
-            form.lsb_order,
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.ByteMaskedForm):
-        return awkward.forms.ByteMaskedForm(
-            form.mask,
-            awkward_form_remove_uproot(awkward, form.content),
-            form.valid_when,
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.EmptyForm):
-        return awkward.forms.EmptyForm(
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.IndexedForm):
-        return awkward.forms.IndexedForm(
-            form.index,
-            awkward_form_remove_uproot(awkward, form.content),
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.IndexedOptionForm):
-        return awkward.forms.IndexedOptionForm(
-            form.index,
-            awkward_form_remove_uproot(awkward, form.content),
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.ListForm):
-        return awkward.forms.ListForm(
-            form.starts,
-            form.stops,
-            awkward_form_remove_uproot(awkward, form.content),
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.ListOffsetForm):
-        return awkward.forms.ListOffsetForm(
-            form.offsets,
-            awkward_form_remove_uproot(awkward, form.content),
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.NumpyForm):
-        return awkward.forms.NumpyForm(
-            form.inner_shape,
-            form.itemsize,
-            form.format,
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.RecordForm):
-        return awkward.forms.RecordForm(
-            {
-                k: awkward_form_remove_uproot(awkward, v)
-                for k, v in form.contents.items()
-                if not k.startswith("@")
-            },
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.RegularForm):
-        return awkward.forms.RegularForm(
-            awkward_form_remove_uproot(awkward, form.content),
-            form.size,
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.UnionForm):
-        return awkward.forms.UnionForm(
-            form.tags,
-            form.index,
-            [awkward_form_remove_uproot(awkward, x) for x in form.contents],
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.UnmaskedForm):
-        return awkward.forms.UnmaskedForm(
-            awkward_form_remove_uproot(awkward, form.content),
-            form.has_identities,
-            parameters,
-        )
-    elif isinstance(form, awkward.forms.VirtualForm):
-        return awkward.forms.VirtualForm(
-            awkward_form_remove_uproot(awkward, form.form),
-            form.has_length,
-            form.has_identities,
-            parameters,
-        )
-    else:
-        raise RuntimeError(f"unrecognized form: {type(form)}")
-
-
-# FIXME: Until we get Awkward reading these bytes directly, rather than
-# going through ak.from_iter, the integer dtypes will be int64 and the
-# floating dtypes will be float64 because that's what ak.from_iter makes.
 def recursively_fix_awkward_form_of_iter(awkward, interpretation, form):
     """
     Given an interpretation of a TBranch, fixup any form corresponding to
@@ -679,9 +575,9 @@ def recursively_fix_awkward_form_of_iter(awkward, interpretation, form):
                 awkward, subinterpretation, form.content(key)
             )
         return awkward.forms.RecordForm(
-            fixed,
-            form.has_identities,
-            form.parameters,
+            list(fixed.values()),
+            list(fixed.keys()),
+            parameters=form.parameters,
         )
     elif isinstance(interpretation, uproot.interpretation.objects.AsObjects):
         if uproot.interpretation.objects.awkward_can_optimize(interpretation, form):
@@ -704,102 +600,86 @@ def awkward_form_of_iter(awkward, form):
             awkward_form_of_iter(awkward, form.content),
             form.valid_when,
             form.lsb_order,
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.ByteMaskedForm):
         return awkward.forms.ByteMaskedForm(
             form.mask,
             awkward_form_of_iter(awkward, form.content),
             form.valid_when,
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.EmptyForm):
         return awkward.forms.EmptyForm(
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.IndexedForm):
         return awkward.forms.IndexedForm(
             form.index,
             awkward_form_of_iter(awkward, form.content),
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.IndexedOptionForm):
         return awkward.forms.IndexedOptionForm(
             form.index,
             awkward_form_of_iter(awkward, form.content),
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.ListForm):
         return awkward.forms.ListForm(
             form.starts,
             form.stops,
             awkward_form_of_iter(awkward, form.content),
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.ListOffsetForm):
         return awkward.forms.ListOffsetForm(
             form.offsets,
             awkward_form_of_iter(awkward, form.content),
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.NumpyForm):
         if form.parameter("__array__") in ("char", "byte"):
             f = form
         else:
-            d = form.to_numpy()
+            d = awkward.types.numpytype.primitive_to_dtype(form.primitive)
             if issubclass(d.type, numpy.integer):
                 d = numpy.dtype(numpy.int64)
             elif issubclass(d.type, numpy.floating):
                 d = numpy.dtype(numpy.float64)
-            f = awkward.forms.Form.from_numpy(d)
+            f = awkward.forms.numpyform.from_dtype(d)
         out = awkward.forms.NumpyForm(
+            f.primitive,
             form.inner_shape,
-            f.itemsize,
-            f.format,
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
         return out
     elif isinstance(form, awkward.forms.RecordForm):
+        contents = {
+            k: awkward_form_of_iter(awkward, v) for k, v in form.contents.items()
+        }
         return awkward.forms.RecordForm(
-            {k: awkward_form_of_iter(awkward, v) for k, v in form.contents.items()},
-            form.has_identities,
-            form.parameters,
+            list(contents.values()),
+            list(contents.keys()),
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.RegularForm):
         return awkward.forms.RegularForm(
             awkward_form_of_iter(awkward, form.content),
             form.size,
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.UnionForm):
         return awkward.forms.UnionForm(
             form.tags,
             form.index,
             [awkward_form_of_iter(awkward, x) for x in form.contents],
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     elif isinstance(form, awkward.forms.UnmaskedForm):
         return awkward.forms.UnmaskedForm(
             awkward_form_of_iter(awkward, form.content),
-            form.has_identities,
-            form.parameters,
-        )
-    elif isinstance(form, awkward.forms.VirtualForm):
-        return awkward.forms.VirtualForm(
-            awkward_form_of_iter(awkward, form.form),
-            form.has_length,
-            form.has_identities,
-            form.parameters,
+            parameters=form.parameters,
         )
     else:
         raise RuntimeError(f"unrecognized form: {type(form)}")
@@ -815,7 +695,7 @@ def damerau_levenshtein(a, b, ratio=False):
     # Modified Damerau-Levenshtein distance. Adds a middling penalty
     # for capitalization.
     # https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance
-    M = [[0] * (len(b) + 1) for i in range(len(a) + 1)]
+    M = [[0] * (len(b) + 1) for _ in range(len(a) + 1)]
 
     for i in range(len(a) + 1):
         M[i][0] = i
@@ -847,7 +727,7 @@ def damerau_levenshtein(a, b, ratio=False):
                     # Transpose only
                     M[i][j] = min(M[i][j], M[i - 2][j - 2] + 1)
                 else:
-                    # Traspose and capitalization
+                    # Transpose and capitalization
                     M[i][j] = min(M[i][j], M[i - 2][j - 2] + 1.5)
 
     if not ratio:
@@ -901,28 +781,75 @@ _regularize_files_braces = re.compile(r"{([^}]*,)*([^}]*)}")
 _regularize_files_isglob = re.compile(r"[\*\?\[\]{}]")
 
 
-def _regularize_files_inner(files, parse_colon, counter, HasBranches):
+def regularize_steps(steps):
+    out = numpy.array(steps)
+
+    if isinstance(steps, dict) or not issubclass(out.dtype.type, numpy.integer):
+        raise TypeError(
+            "'files' argument's steps must be an iterable of integer offsets or start-stop pairs."
+        )
+
+    if len(out.shape) == 1:
+        if len(out) == 0 or not numpy.all(out[1:] >= out[:-1]):
+            raise ValueError(
+                "if 'files' argument's steps are (one-dimensional) offsets, they must be non-empty and monotonically increasing"
+            )
+
+    elif len(out.shape) == 2:
+        if not (out.shape[1] == 2 and all(out[:, 1] >= out[:, 0])):
+            raise ValueError(
+                "if 'files' argument's steps are (two-dimensional) start-stop pairs, all stops must be greater than or equal to their corresponding starts"
+            )
+
+    else:
+        raise TypeError(
+            "'files' argument's steps must be an iterable of integer offsets or a list of pairs of integer starts and stops."
+        )
+
+    if len(out.shape) == 1:
+        out = numpy.stack((out[:-1], out[1:]), axis=1)
+
+    return out.tolist()
+
+
+def _regularize_files_inner(
+    files, parse_colon, counter, HasBranches, steps_allowed, **options
+):
     files2 = regularize_path(files)
 
-    if isstr(files2) and not isstr(files):
+    maybe_steps = None
+
+    if isinstance(files2, str) and not isinstance(files, str):
         parse_colon = False
         files = files2
 
-    if isstr(files):
+    if isinstance(files, str):
         if parse_colon:
             file_path, object_path = file_object_path_split(files)
         else:
             file_path, object_path = files, None
 
+        # This parses the windows drive letter as a scheme!
         parsed_url = urlparse(file_path)
-
-        if parsed_url.scheme.upper() in _remote_schemes:
-            yield file_path, object_path
-
+        scheme = parsed_url.scheme
+        if "://" in file_path and scheme not in ("file", "local"):
+            # user specified a protocol, so we use fsspec to expand the glob and return the full paths
+            file_names_full = [
+                file.full_name
+                for file in fsspec.open_files(
+                    file_path,
+                    **uproot.source.fsspec.FSSpecSource.extract_fsspec_options(options),
+                )
+            ]
+            # https://github.com/fsspec/filesystem_spec/issues/1459
+            # Not all protocols return the full_name attribute correctly (if they have url parameters)
+            for file_name_full in file_names_full:
+                yield file_name_full, object_path, maybe_steps
         else:
+            # no protocol, default to local file system
             expanded = os.path.expanduser(file_path)
             if _regularize_files_isglob.search(expanded) is None:
-                yield file_path, object_path
+                yield file_path, object_path, maybe_steps
 
             else:
                 matches = list(_regularize_files_braces.finditer(expanded))
@@ -942,37 +869,55 @@ def _regularize_files_inner(files, parse_colon, counter, HasBranches):
                 for result in results:
                     for match in glob.glob(result):
                         if match not in seen:
-                            yield match, object_path
+                            yield match, object_path, maybe_steps
                             seen.add(match)
 
     elif isinstance(files, HasBranches):
-        yield files, None
+        yield files, None, maybe_steps
 
     elif isinstance(files, dict):
-        for key, object_path in files.items():
-            for file_path, _ in _regularize_files_inner(
-                key, False, counter, HasBranches
+        for key, maybe_object_path in files.items():
+            if not isinstance(maybe_object_path, (type(None), str, dict)):
+                raise TypeError("object_path may only be a string, dict, or None")
+            if isinstance(maybe_object_path, dict):
+                maybe_steps = maybe_object_path.get("steps", None)
+                object_path = maybe_object_path.get("object_path", None)
+                if maybe_steps is not None:
+                    if not steps_allowed:
+                        raise TypeError(
+                            "unrecognized 'files' pattern for this function ('steps' are only allowed in uproot.dask)"
+                        )
+                    maybe_steps = regularize_steps(maybe_steps)
+            else:
+                object_path = maybe_object_path
+            for file_path, _, _ in _regularize_files_inner(
+                key,
+                False,
+                counter,
+                HasBranches,
+                steps_allowed,
+                **options,
             ):
-                yield file_path, object_path
+                yield file_path, object_path, maybe_steps
 
     elif isinstance(files, Iterable):
         for file in files:
             counter[0] += 1
-            for file_path, object_path in _regularize_files_inner(
-                file, parse_colon, counter, HasBranches
+            for file_path, object_path, maybe_steps in _regularize_files_inner(
+                file, parse_colon, counter, HasBranches, steps_allowed, **options
             ):
-                yield file_path, object_path
+                yield file_path, object_path, maybe_steps
 
     else:
         raise TypeError(
             "'files' must be a file path/URL (string or Path), possibly with "
             "a glob pattern (for local files), a dict of "
-            "{{path/URL: TTree/TBranch name}}, actual TTree/TBranch objects, or "
-            "an iterable of such things, not {0}".format(repr(files))
+            "{path/URL: TTree/TBranch name}, actual TTree/TBranch objects, or "
+            f"an iterable of such things, not {files!r}"
         )
 
 
-def regularize_files(files):
+def regularize_files(files, steps_allowed, **options):
     """
     Common code for regularizing the possible file inputs accepted by uproot so they can be used by uproot internal functions.
     """
@@ -981,16 +926,21 @@ def regularize_files(files):
     out = []
     seen = set()
     counter = [0]
-    for file_path, object_path in _regularize_files_inner(
-        files, True, counter, HasBranches
+    for file_path, object_path, maybe_steps in _regularize_files_inner(
+        files, True, counter, HasBranches, steps_allowed, **options
     ):
-        if isstr(file_path):
+        if isinstance(file_path, str):
             key = (counter[0], file_path, object_path)
             if key not in seen:
                 out.append((file_path, object_path))
+                if maybe_steps is not None:
+                    out[-1] = (*out[-1], maybe_steps)
+
                 seen.add(key)
         else:
             out.append((file_path, object_path))
+            if maybe_steps is not None:
+                out[-1] = (*out[-1], maybe_steps)
 
     if len(out) == 0:
         raise _file_not_found(files)
@@ -1016,7 +966,7 @@ def regularize_object_path(
             object_cache=None,
             array_cache=None,
             custom_classes=custom_classes,
-            **options,  # NOTE: a comma after **options breaks Python 2
+            **options,
         ).root_directory
         if object_path is None:
             trees = file.keys(filter_classname="TTree", cycle=False)
@@ -1024,28 +974,53 @@ def regularize_object_path(
                 if allow_missing:
                     return None
                 else:
-                    raise ValueError(
-                        """no TTrees found
-in file {}""".format(
-                            file_path
-                        )
-                    )
+                    raise ValueError(f"no TTrees found\nin file {file_path}")
             elif len(trees) == 1:
                 return file[trees[0]]
             else:
+                ttree_str = ", ".join(repr(x) for x in trees)
                 raise ValueError(
                     """TTree object paths must be specified in the 'files' """
-                    """as {{\"filenames*.root\": \"path\"}} if any files have """
-                    """more than one TTree
+                    """as {"filenames*.root": "path"} if any files have """
+                    f"""more than one TTree
 
-    TTrees: {0}
+    TTrees: {ttree_str}
 
-in file {1}""".format(
-                        ", ".join(repr(x) for x in trees), file_path
-                    )
+in file {file_path}"""
                 )
 
         else:
             if allow_missing and object_path not in file:
                 return None
             return file[object_path]
+
+
+def _content_cls_from_name(awkward, name):
+    if name.endswith(("32", "64")):
+        name = name[-2:]
+    elif name.endswith("U32"):
+        name = name[-3:]
+    elif name.endswith(("8_32", "8_64")):
+        name = name[-4:]
+    elif name.endswith("8_U32"):
+        name = name[-5:]
+    return getattr(awkward.contents, name)
+
+
+def pandas_has_attr_is_numeric(pandas):
+    try:
+        function = pandas.api.types.is_any_real_numeric_dtype
+    except AttributeError:
+
+        def function(x):
+            return x.is_numeric
+
+    return function
+
+
+class _Unset:
+    def __repr__(self):
+        return f"{__name__}.unset"
+
+
+unset = _Unset()
