@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines utilities for identifying the
@@ -11,9 +11,10 @@ need to be tweaked by new types, type combinations, and serialization methods
 observed in ROOT files (perhaps forever), unless a systematic study can be
 performed to exhaustively discover all cases.
 """
-
+from __future__ import annotations
 
 import ast
+import numbers
 import re
 
 import numpy
@@ -84,6 +85,11 @@ def _leaf_to_dtype(leaf, getdims):
             return numpy.dtype((numpy.uint32, dims))
         else:
             return numpy.dtype((numpy.int32, dims))
+    elif leaf.classname == "TLeafG":
+        if leaf.member("fIsUnsigned"):
+            return numpy.dtype((numpy.uint64, dims))
+        else:
+            return numpy.dtype((numpy.int64, dims))
     elif leaf.classname == "TLeafL":
         if leaf.member("fIsUnsigned"):
             return numpy.dtype((numpy.uint64, dims))
@@ -110,9 +116,8 @@ def _from_leaves_one(leaf, title):
     m = _title_has_dims.match(title)
     if m is not None:
         dims = tuple(int(x) for x in re.findall(_item_dim_pattern, title))
-        if dims == ():
-            if leaf.member("fLen") > 1:
-                dims = (leaf.member("fLen"),)
+        if dims == () and leaf.member("fLen") > 1:
+            dims = (leaf.member("fLen"),)
 
         if any(
             _item_dim_pattern.match(x) is None
@@ -160,15 +165,15 @@ def _float16_double32_walk_ast(node, branch, source):
             and isinstance(node.ctx, ast.Load)
             and node.id.lower() == "pi"
         ):
-            out = ast.Num(3.141592653589793)  # TMath::Pi()
+            out = ast.Constant(3.141592653589793)  # TMath::Pi()
         elif (
             isinstance(node, ast.Name)
             and isinstance(node.ctx, ast.Load)
             and node.id.lower() == "twopi"
         ):
-            out = ast.Num(6.283185307179586)  # TMath::TwoPi()
-        elif isinstance(node, ast.Num):
-            out = ast.Num(float(node.n))
+            out = ast.Constant(6.283185307179586)  # TMath::TwoPi()
+        elif isinstance(node, ast.Constant) and isinstance(node.value, numbers.Number):
+            out = ast.Constant(float(node.value))
         elif isinstance(node, ast.BinOp) and isinstance(
             node.op, (ast.Add, ast.Sub, ast.Mult, ast.Div)
         ):
@@ -197,7 +202,8 @@ def _float16_double32_walk_ast(node, branch, source):
             isinstance(node, ast.List)
             and isinstance(node.ctx, ast.Load)
             and len(node.elts) == 3
-            and isinstance(node.elts[2], ast.Num)
+            and isinstance(node.elts[2], ast.Constant)
+            and isinstance(node.elts[2].value, numbers.Number)
         ):
             out = ast.List(
                 [
@@ -476,15 +482,23 @@ def interpretation_of(branch, context, simplify=True):
                 while isinstance(model_cls, uproot.containers.AsPointer):
                     model_cls = model_cls.pointee
 
-            if branch._streamer_isTClonesArray:
-                if isinstance(branch.streamer, uproot.streamers.Model_TStreamerObject):
+            if dims != () or branch._streamer_isTClonesArray:
+                if (
+                    isinstance(branch.streamer, uproot.streamers.Model_TStreamerObject)
+                    or not branch._streamer_isTClonesArray
+                ):
                     model_cls = uproot.containers.AsArray(False, False, model_cls, dims)
                 else:
                     if hasattr(model_cls, "header"):
                         model_cls._header = False
                     model_cls = uproot.containers.AsArray(True, False, model_cls, dims)
 
-            out = uproot.interpretation.objects.AsObjects(model_cls, branch)
+            if dims != () and not branch._streamer_isTClonesArray:
+                out = uproot.interpretation.objects.AsObjects(model_cls, branch)
+                out._forth = False
+            else:
+                out = uproot.interpretation.objects.AsObjects(model_cls, branch)
+
             if simplify:
                 return out.simplify()
             else:
@@ -508,10 +522,7 @@ _simplify_token_4 = re.compile(r"\s*>\s*")
 
 
 def _simplify_token(token, is_token=True):
-    if is_token:
-        text = token.group(0)
-    else:
-        text = token
+    text = token.group(0) if is_token else token
     text = _simplify_token_1.sub("*", text)
     text = _simplify_token_2.sub("::", text)
     text = _simplify_token_3.sub("<", text)
@@ -601,9 +612,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype("u1"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype("u1"))',
                 quote,
             ),
         )
@@ -615,9 +624,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 2,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype("u1"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype("u1"))',
                 quote,
             ),
         )
@@ -637,9 +644,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">i2"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">i2"))',
                 quote,
             ),
         )
@@ -647,9 +652,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">i2"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">i2"))',
                 quote,
             ),
         )
@@ -657,9 +660,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">u2"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">u2"))',
                 quote,
             ),
         )
@@ -671,9 +672,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 2,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">u2"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">u2"))',
                 quote,
             ),
         )
@@ -691,9 +690,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">i4"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">i4"))',
                 quote,
             ),
         )
@@ -701,9 +698,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">i4"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">i4"))',
                 quote,
             ),
         )
@@ -711,9 +706,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">u4"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">u4"))',
                 quote,
             ),
         )
@@ -725,20 +718,17 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 2,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">u4"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">u4"))',
                 quote,
             ),
         )
 
-    elif has2 and tokens[i].group(0) == "long" and tokens[i + 1].group(0) == "long":
+    elif has2 and tokens[i].group(0) == tokens[i + 1].group(0) == "long":
         return i + 2, _parse_maybe_quote('numpy.dtype(">i8")', quote)
     elif (
         i + 2 < len(tokens)
         and tokens[i].group(0) == "unsigned"
-        and tokens[i + 1].group(0) == "long"
-        and tokens[i + 2].group(0) == "long"
+        and tokens[i + 1].group(0) == tokens[i + 2].group(0) == "long"
     ):
         return i + 3, _parse_maybe_quote('numpy.dtype(">u8")', quote)
 
@@ -785,9 +775,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">i8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">i8"))',
                 quote,
             ),
         )
@@ -795,9 +783,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">i8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">i8"))',
                 quote,
             ),
         )
@@ -805,9 +791,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">i8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">i8"))',
                 quote,
             ),
         )
@@ -815,9 +799,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">u8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">u8"))',
                 quote,
             ),
         )
@@ -825,9 +807,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">u8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">u8"))',
                 quote,
             ),
         )
@@ -839,9 +819,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 2,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">u8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">u8"))',
                 quote,
             ),
         )
@@ -855,9 +833,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">f4"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">f4"))',
                 quote,
             ),
         )
@@ -865,9 +841,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">f4"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">f4"))',
                 quote,
             ),
         )
@@ -881,9 +855,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">f8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">f8"))',
                 quote,
             ),
         )
@@ -891,9 +863,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                'uproot.containers.AsArray(False, {}, numpy.dtype(">f8"))'.format(
-                    header
-                ),
+                f'uproot.containers.AsArray(False, {header}, numpy.dtype(">f8"))',
                 quote,
             ),
         )
@@ -910,8 +880,8 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                "uproot.containers.AsArray(False, {}, "
-                'uproot.containers.AsFIXME("Float16_t in array"))'.format(header),
+                f"uproot.containers.AsArray(False, {header}, "
+                'uproot.containers.AsFIXME("Float16_t in array"))',
                 quote,
             ),
         )
@@ -928,9 +898,9 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 1,
             _parse_maybe_quote(
-                "uproot.containers.AsArray(False, {}, "
+                f"uproot.containers.AsArray(False, {header}, "
                 'uproot.containers.AsFIXME("Double32_t in array '
-                '(note: Event.root fClosestDistance has an example)"))'.format(header),
+                '(note: Event.root fClosestDistance has an example)"))',
                 quote,
             ),
         )
@@ -940,6 +910,20 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
             i + 1,
             _parse_maybe_quote(f"uproot.containers.AsString({header})", quote),
         )
+
+    elif (
+        has2
+        and tokens[i].group(0) == "const"
+        and (
+            tokens[i + 1].group(0) == "string"
+            or _simplify_token(tokens[i + 1]) == "std::string"
+        )
+    ):
+        return (
+            i + 2,
+            _parse_maybe_quote(f"uproot.containers.AsString({header})", quote),
+        )
+
     elif tokens[i].group(0) == "TString":
         return (
             i + 1,
@@ -980,7 +964,7 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         return (
             i + 4,
             _parse_maybe_quote(
-                f'uproot.containers.AsFIXME("std::bitset<{num_bits}>")',
+                f"uproot.containers.AsBitSet({header}, {num_bits})",
                 quote,
             ),
         )
@@ -1000,6 +984,21 @@ def _parse_node(tokens, i, typename, file, quote, header, inner_header):
         else:
             return i + 1, uproot.containers.AsVector(header, values)
 
+    elif tokens[i].group(0) == "list" or _simplify_token(tokens[i]) == "std::list":
+        _parse_expect("<", tokens, i + 1, typename, file)
+        i, values = _parse_node(
+            tokens, i + 2, typename, file, quote, inner_header, inner_header
+        )
+        i = _parse_ignore_extra_arguments(tokens, i, typename, file, 1)
+        _parse_expect(">", tokens, i, typename, file)
+        if quote:
+            return (
+                i + 1,
+                f"uproot.containers.AsList({header}, {values})",
+            )
+        else:
+            return i + 1, uproot.containers.AsList(header, values)
+
     elif (
         tokens[i].group(0) == "RVec"
         or _simplify_token(tokens[i]) == "VecOps::RVec"
@@ -1144,8 +1143,6 @@ class NotNumerical(Exception):
     non-conforming type is found.
     """
 
-    pass
-
 
 class UnknownInterpretation(Exception):
     """
@@ -1170,11 +1167,9 @@ class UnknownInterpretation(Exception):
         return f"<UnknownInterpretation {self.reason!r}>"
 
     def __str__(self):
-        return """{}
-in file {}
-in object {}""".format(
-            self.reason, self.file_path, self.object_path
-        )
+        return f"""{self.reason}
+in file {self.file_path}
+in object {self.object_path}"""
 
     @property
     def typename(self):
@@ -1194,7 +1189,7 @@ in object {}""".format(
         context=None,
         index_format="i64",
         header=False,
-        tobject_header=True,
+        tobject_header=False,
         breadcrumbs=(),
     ):
         raise self
