@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines an :doc:`uproot.interpretation.Interpretation` for
@@ -15,7 +15,7 @@ several numerical types:
 * :doc:`uproot.interpretation.numerical.AsSTLBits`: an ``std::bitset<N>``
   for some ``N``.
 """
-
+from __future__ import annotations
 
 import numpy
 
@@ -43,7 +43,14 @@ class Numerical(uproot.interpretation.Interpretation):
         return array
 
     def final_array(
-        self, basket_arrays, entry_start, entry_stop, entry_offsets, library, branch
+        self,
+        basket_arrays,
+        entry_start,
+        entry_stop,
+        entry_offsets,
+        library,
+        branch,
+        options,
     ):
         self.hook_before_final_array(
             basket_arrays=basket_arrays,
@@ -99,6 +106,7 @@ class Numerical(uproot.interpretation.Interpretation):
 
                 start = stop
 
+        output = output.view(output.dtype.newbyteorder("="))
         self.hook_before_library_finalize(
             basket_arrays=basket_arrays,
             entry_start=entry_start,
@@ -111,7 +119,9 @@ class Numerical(uproot.interpretation.Interpretation):
 
         output = self._wrap_almost_finalized(output)
 
-        output = library.finalize(output, branch, self, entry_start, entry_stop)
+        output = library.finalize(
+            output, branch, self, entry_start, entry_stop, options
+        )
 
         self.hook_after_final_array(
             basket_arrays=basket_arrays,
@@ -182,11 +192,9 @@ class AsDtype(Numerical):
 
     def __repr__(self):
         if self._to_dtype == self._from_dtype.newbyteorder("="):
-            return f"AsDtype({repr(str(self._from_dtype))})"
+            return f"AsDtype({str(self._from_dtype)!r})"
         else:
-            return "AsDtype({}, {})".format(
-                repr(str(self._from_dtype)), repr(str(self._to_dtype))
-            )
+            return f"AsDtype({str(self._from_dtype)!r}, {str(self._to_dtype)!r})"
 
     def __eq__(self, other):
         return (
@@ -255,7 +263,7 @@ class AsDtype(Numerical):
         context=None,
         index_format="i64",
         header=False,
-        tobject_header=True,
+        tobject_header=False,
         breadcrumbs=(),
     ):
         context = self._make_context(
@@ -325,7 +333,15 @@ class AsDtype(Numerical):
             )
 
     def basket_array(
-        self, data, byte_offsets, basket, branch, context, cursor_offset, library
+        self,
+        data,
+        byte_offsets,
+        basket,
+        branch,
+        context,
+        cursor_offset,
+        library,
+        options,
     ):
         self.hook_before_basket_array(
             data=data,
@@ -335,22 +351,17 @@ class AsDtype(Numerical):
             context=context,
             cursor_offset=cursor_offset,
             library=library,
+            options=options,
         )
 
         dtype, shape = _dtype_shape(self._from_dtype)
         try:
-            output = data.view(dtype).reshape((-1,) + shape)
+            output = data.view(dtype).reshape((-1, *shape))
         except ValueError as err:
             raise ValueError(
-                """basket {} in tree/branch {} has the wrong number of bytes ({}) """
-                """for interpretation {}
-in file {}""".format(
-                    basket.basket_num,
-                    branch.object_path,
-                    len(data),
-                    self,
-                    branch.file.file_path,
-                )
+                f"""basket {basket.basket_num} in tree/branch {branch.object_path} has the wrong number of bytes ({len(data)}) """
+                f"""for interpretation {self}
+in file {branch.file.file_path}"""
             ) from err
 
         self.hook_after_basket_array(
@@ -362,6 +373,7 @@ in file {}""".format(
             output=output,
             cursor_offset=cursor_offset,
             library=library,
+            options=options,
         )
 
         return output
@@ -403,18 +415,14 @@ class AsDtypeInPlace(AsDtype):
         """
         if library.name != "np":
             raise TypeError(
-                "AsDtypeInPlace can only be used with library 'np', not '{}'".format(
-                    library.name
-                )
+                f"AsDtypeInPlace can only be used with library 'np', not '{library.name}'"
             )
 
         output = self._to_fill.view(self.to_dtype)
 
         if length > len(output):
             raise ValueError(
-                "Requesting to fill an array of size {} (type {}) with input of size {} (type {})".format(
-                    len(output), self._to_dtype, length, self._from_dtype
-                )
+                f"Requesting to fill an array of size {len(output)} (type {self._to_dtype}) with input of size {length} (type {self._from_dtype})"
             )
         return output[:length]
 
@@ -500,7 +508,7 @@ class TruncatedNumerical(Numerical):
         and :ref:`uproot.interpretation.numerical.TruncatedNumerical.high` are
         both ``0``), the data are truly truncated.
         """
-        return self._low == 0.0 and self._high == 0.0
+        return self._low == self._high == 0.0
 
     def __repr__(self):
         args = [repr(self._low), repr(self._high), repr(self._num_bits)]
@@ -523,12 +531,18 @@ class TruncatedNumerical(Numerical):
 
     @property
     def cache_key(self):
-        return "{}({},{},{},{})".format(
-            type(self).__name__, self._low, self._high, self._num_bits, self._to_dims
-        )
+        return f"{type(self).__name__}({self._low},{self._high},{self._num_bits},{self._to_dims})"
 
     def basket_array(
-        self, data, byte_offsets, basket, branch, context, cursor_offset, library
+        self,
+        data,
+        byte_offsets,
+        basket,
+        branch,
+        context,
+        cursor_offset,
+        library,
+        options,
     ):
         self.hook_before_basket_array(
             data=data,
@@ -538,22 +552,16 @@ class TruncatedNumerical(Numerical):
             context=context,
             cursor_offset=cursor_offset,
             library=library,
+            options=options,
         )
 
         try:
             raw = data.view(self.from_dtype)
         except ValueError as err:
             raise ValueError(
-                """basket {} in tree/branch {} has the wrong number of bytes ({}) """
-                """for interpretation {} (expecting raw array of {})
-in file {}""".format(
-                    basket.basket_num,
-                    branch.object_path,
-                    len(data),
-                    self,
-                    repr(self._from_dtype),
-                    branch.file.file_path,
-                )
+                f"""basket {basket.basket_num} in tree/branch {branch.object_path} has the wrong number of bytes ({len(data)}) """
+                f"""for interpretation {self} (expecting raw array of {self._from_dtype!r})
+in file {branch.file.file_path}"""
             ) from err
 
         if self.is_truncated:
@@ -569,11 +577,11 @@ in file {}""".format(
             output = exponent.view(numpy.float32) * sign
 
             d, s = _dtype_shape(self.to_dtype)
-            output = output.astype(d).reshape((-1,) + s)
+            output = output.astype(d).reshape((-1, *s))
 
         else:
             d, s = _dtype_shape(self.to_dtype)
-            output = raw.astype(d).reshape((-1,) + s)
+            output = raw.astype(d).reshape((-1, *s))
             numpy.multiply(
                 output,
                 float(self._high - self._low) / (1 << self._num_bits),
@@ -591,6 +599,7 @@ in file {}""".format(
             library=library,
             raw=raw,
             output=output,
+            options=options,
         )
 
         return output
@@ -641,26 +650,14 @@ class AsDouble32(TruncatedNumerical):
         context=None,
         index_format="i64",
         header=False,
-        tobject_header=True,
+        tobject_header=False,
         breadcrumbs=(),
     ):
         context = self._make_context(
             context, index_format, header, tobject_header, breadcrumbs
         )
         awkward = uproot.extras.awkward()
-        out = awkward.forms.NumpyForm(
-            (),
-            8,
-            "d",
-            parameters={
-                "uproot": {
-                    "as": "Double32",
-                    "low": self._low,
-                    "high": self._high,
-                    "num_bits": self._num_bits,
-                }
-            },
-        )
+        out = awkward.forms.NumpyForm("float64")
         for size in self._to_dims[::-1]:
             out = awkward.forms.RegularForm(out, size)
         return out
@@ -711,26 +708,14 @@ class AsFloat16(TruncatedNumerical):
         context=None,
         index_format="i64",
         header=False,
-        tobject_header=True,
+        tobject_header=False,
         breadcrumbs=(),
     ):
         context = self._make_context(
             context, index_format, header, tobject_header, breadcrumbs
         )
         awkward = uproot.extras.awkward()
-        out = awkward.forms.NumpyForm(
-            (),
-            4,
-            "f",
-            parameters={
-                "uproot": {
-                    "as": "Float16",
-                    "low": self._low,
-                    "high": self._high,
-                    "num_bits": self._num_bits,
-                }
-            },
-        )
+        out = awkward.forms.NumpyForm("float32")
         for size in self._to_dims[::-1]:
             out = awkward.forms.RegularForm(out, size)
         return out
