@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines an :doc:`uproot.interpretation.Interpretation` and
@@ -18,28 +18,16 @@ The :doc:`uproot.interpretation.objects.ObjectArray` and
 while an array is being built from ``TBaskets``. Its final form is determined
 by the :doc:`uproot.interpretation.library.Library`.
 """
+from __future__ import annotations
 
+import contextlib
+import json
+import threading
 
 import numpy
 
 import uproot
-
-
-def awkward_can_optimize(interpretation, form):
-    """
-    If True, the Awkward Array library can convert data of a given
-    :doc:`uproot.interpretation.Interpretation` and ``ak.forms.Form`` into
-    arrays without resorting to ``ak.from_iter`` (i.e. rapidly).
-
-    If ``awkward._connect._uproot`` cannot be imported, this function always
-    returns False.
-    """
-    try:
-        import awkward._connect._uproot
-    except ModuleNotFoundError:
-        return False
-    else:
-        return awkward._connect._uproot.can_optimize(interpretation, form)
+import uproot._awkwardforth
 
 
 class AsObjects(uproot.interpretation.Interpretation):
@@ -62,6 +50,10 @@ class AsObjects(uproot.interpretation.Interpretation):
     def __init__(self, model, branch=None):
         self._model = model
         self._branch = branch
+        self._form = None
+        self._forth = True
+        self._complete_forth_code = None
+        self._forth_lock = threading.Lock()
 
     @property
     def model(self):
@@ -110,19 +102,24 @@ class AsObjects(uproot.interpretation.Interpretation):
         context=None,
         index_format="i64",
         header=False,
-        tobject_header=True,
+        tobject_header=False,
         breadcrumbs=(),
     ):
         context = self._make_context(
             context, index_format, header, tobject_header, breadcrumbs
         )
-        if isinstance(self._model, type):
-            return self._model.awkward_form(self._branch.file, context)
-        else:
-            return self._model.awkward_form(self._branch.file, context)
+        return self._model.awkward_form(self._branch.file, context)
 
     def basket_array(
-        self, data, byte_offsets, basket, branch, context, cursor_offset, library
+        self,
+        data,
+        byte_offsets,
+        basket,
+        branch,
+        context,
+        cursor_offset,
+        library,
+        options,
     ):
         self.hook_before_basket_array(
             data=data,
@@ -132,28 +129,31 @@ class AsObjects(uproot.interpretation.Interpretation):
             context=context,
             cursor_offset=cursor_offset,
             library=library,
+            options=options,
         )
         assert basket.byte_offsets is not None
 
-        output = None
-        if isinstance(library, uproot.interpretation.library.Awkward):
-            form = self.awkward_form(branch.file)
+        if self._forth and (
+            isinstance(
+                library,
+                (
+                    uproot.interpretation.library.Awkward,
+                    uproot.interpretation.library.Pandas,
+                ),
+            )
+        ):
+            output = self.basket_array_forth(
+                data,
+                byte_offsets,
+                basket,
+                branch,
+                context,
+                cursor_offset,
+                library,
+                options,
+            )
 
-            if awkward_can_optimize(self, form):
-                import awkward._connect._uproot
-
-                extra = {
-                    "interpretation": self,
-                    "basket": basket,
-                    "branch": branch,
-                    "context": context,
-                    "cursor_offset": cursor_offset,
-                }
-                output = awkward._connect._uproot.basket_array(
-                    form, data, byte_offsets, extra
-                )
-
-        if output is None:
+        else:
             output = ObjectArray(
                 self._model, branch, context, byte_offsets, data, cursor_offset
             ).to_numpy()
@@ -167,12 +167,264 @@ class AsObjects(uproot.interpretation.Interpretation):
             output=output,
             cursor_offset=cursor_offset,
             library=library,
+            options=options,
         )
 
         return output
 
+    def basket_array_forth(
+        self,
+        data,
+        byte_offsets,
+        basket,
+        branch,
+        context,
+        cursor_offset,
+        library,
+        options,
+    ):
+        awkward = uproot.extras.awkward()
+        import awkward.forth
+
+        self.hook_before_basket_array(
+            data=data,
+            byte_offsets=byte_offsets,
+            basket=basket,
+            branch=branch,
+            context=context,
+            cursor_offset=cursor_offset,
+            library=library,
+            options=options,
+        )
+        assert basket.byte_offsets is not None
+
+        if getattr(context["forth"], "vm", None) is None:
+            # context["forth"] is a threading.local()
+            context["forth"].vm = None
+            context["forth"].gen = uproot._awkwardforth.ForthGenerator(self)
+        else:
+            context["forth"].gen = None
+
+        if self._complete_forth_code is None:
+            # all threads have to wait until complete_forth_code is ready
+            with self._forth_lock:
+                # another thread might have concluded that this can't be Forth
+                if not self._forth:
+                    return self.basket_array(
+                        data,
+                        byte_offsets,
+                        basket,
+                        branch,
+                        context,
+                        cursor_offset,
+                        library,
+                        options,
+                    )
+
+                # another thread might have already found the complete Forth code
+                elif self._complete_forth_code is None:
+                    context = dict(context)
+                    context["path"] = ()
+
+                    # this thread tries to do it!
+                    try:
+                        output = self._discover_forth(
+                            data, byte_offsets, branch, context, cursor_offset
+                        )
+                    except CannotBeForth:
+                        self._forth = False
+                        context["forth"].gen = None
+                        context["forth"].vm = None
+                        return self.basket_array(
+                            data,
+                            byte_offsets,
+                            basket,
+                            branch,
+                            context,
+                            cursor_offset,
+                            library,
+                            options,
+                        )
+
+                    # Forth discovery was unsuccessful; return Python-derived
+                    # output and maybe another basket will succeed
+                    if output is not None:
+                        self.hook_after_basket_array(
+                            data=data,
+                            byte_offsets=byte_offsets,
+                            basket=basket,
+                            branch=branch,
+                            context=context,
+                            output=output,
+                            cursor_offset=cursor_offset,
+                            library=library,
+                            options=options,
+                        )
+                        return output
+
+        # if we didn't return already, either this thread generated the
+        # complete_forth_code or another did
+        assert self._complete_forth_code is not None
+
+        if context["forth"].vm is None:
+            try:
+                context["forth"].vm = awkward.forth.ForthMachine64(
+                    self._complete_forth_code
+                )
+            except Exception as err:
+                raise type(err)(
+                    str(err)
+                    + "\n\nForth code generated for this data type:\n\n"
+                    + self._complete_forth_code
+                ) from err
+
+        # get data using Forth
+        temp_data = data[byte_offsets[0] : byte_offsets[-1]]
+        context["forth"].vm.begin(
+            {
+                "stream": numpy.array(temp_data),
+                "byteoffsets": numpy.array(byte_offsets[:-1]),
+                "bytestops": numpy.array(byte_offsets[1:]),
+            }
+        )
+        context["forth"].vm.stack_push(len(byte_offsets) - 1)
+        context["forth"].vm.resume()
+        container = {}
+        container = context["forth"].vm.outputs
+        output = awkward.from_buffers(self._form, len(byte_offsets) - 1, container)
+
+        self.hook_after_basket_array(
+            data=data,
+            byte_offsets=byte_offsets,
+            basket=basket,
+            branch=branch,
+            context=context,
+            output=output,
+            cursor_offset=cursor_offset,
+            library=library,
+            options=options,
+        )
+
+        return output
+
+    def _assemble_forth(self, forth_obj, node, only_base=False):
+        if only_base != node.name.startswith("base-class "):
+            return
+
+        forth_obj.final_header.extend(node.header_code)
+        forth_obj.final_init.extend(node.init_code)
+
+        if node.name.startswith("read-members "):
+            for child in node.children:
+                self._assemble_forth(forth_obj, child, only_base=True)
+        forth_obj.final_code.extend(node.pre_code)
+        for child in node.children:
+            self._assemble_forth(forth_obj, child)
+
+        forth_obj.final_code.extend(node.post_code)
+
+    def _any_NULL(self, form):
+        # Recursion through form.
+        # If NULL is found anywhere, return False as Forth is not fully discovered.
+        if form == "NULL":
+            return True
+        else:
+            content = form.get("content")
+            if content is not None:
+                return self._any_NULL(content)
+            contents = form.get("contents")
+            if isinstance(contents, list):
+                for content in contents:
+                    if self._any_NULL(content):
+                        return True
+            elif isinstance(contents, dict):
+                for content in contents.values():
+                    if self._any_NULL(content):
+                        return True
+            return False
+
+    def _debug_forth(self, forth_obj):
+        self._assemble_forth(forth_obj, forth_obj.model.children[0])
+        expected_form = self._model.awkward_form(
+            self._branch.file,
+            {
+                "index_format": "i64",
+                "header": False,
+                "tobject_header": False,
+                "breadcrumbs": (),
+            },
+        )
+
+        raise Exception(
+            f"""EXPECTED FORM:
+{expected_form}
+
+DISCOVERED FORM:
+{json.dumps(forth_obj.model.derive_form(), indent=4)}
+
+FORTH CODE:
+input stream
+    input byteoffsets
+    input bytestops
+    {"".join(forth_obj.final_header)}
+    {"".join(forth_obj.final_init)}
+    0 do
+    byteoffsets I-> stack
+    stream seek
+    {"".join(forth_obj.final_code)}
+    loop
+    """
+        )
+
+    def _discover_forth(self, data, byte_offsets, branch, context, cursor_offset):
+        output = numpy.empty(len(byte_offsets) - 1, dtype=numpy.dtype(object))
+
+        for i in range(len(byte_offsets) - 1):
+            chunk = uproot.source.chunk.Chunk.wrap(
+                branch.file.source, data[byte_offsets[i] : byte_offsets[i + 1]]
+            )
+            cursor = uproot.source.cursor.Cursor(
+                0, origin=-(byte_offsets[i] + cursor_offset)
+            )
+
+            context["forth"].gen.reset_active_node()
+            output[i] = self._model.read(
+                chunk, cursor, context, branch.file, branch.file.detached, branch
+            )
+
+            derived_form = context["forth"].gen.model.derive_form()
+            if not self._any_NULL(derived_form):
+                context["forth"].prereaddone = True
+                self._assemble_forth(
+                    context["forth"].gen, context["forth"].gen.model.children[0]
+                )
+                self._complete_forth_code = f"""input stream
+    input byteoffsets
+    input bytestops
+    {"".join(context["forth"].gen.final_header)}
+    {"".join(context["forth"].gen.final_init)}
+    0 do
+    byteoffsets I-> stack
+    stream seek
+    {"".join(context["forth"].gen.final_code)}
+    loop
+    """
+                self._form = derived_form
+
+                return None  # we should re-read all the data with Forth
+
+        # Python output when Forth-Generation not successful
+        return output
+
     def final_array(
-        self, basket_arrays, entry_start, entry_stop, entry_offsets, library, branch
+        self,
+        basket_arrays,
+        entry_start,
+        entry_stop,
+        entry_offsets,
+        library,
+        branch,
+        options,
     ):
         self.hook_before_final_array(
             basket_arrays=basket_arrays,
@@ -181,33 +433,81 @@ class AsObjects(uproot.interpretation.Interpretation):
             entry_offsets=entry_offsets,
             library=library,
             branch=branch,
+            options=options,
         )
         trimmed = []
         start = entry_offsets[0]
+        has_any_awkward_types = any(
+            uproot._util.from_module(x, "awkward") for x in basket_arrays.values()
+        )
         for basket_num, stop in enumerate(entry_offsets[1:]):
+            to_append = None
             if start <= entry_start and entry_stop <= stop:
                 local_start = entry_start - start
                 local_stop = entry_stop - start
-                trimmed.append(basket_arrays[basket_num][local_start:local_stop])
+                to_append = basket_arrays[basket_num][local_start:local_stop]
 
             elif start <= entry_start < stop:
                 local_start = entry_start - start
                 local_stop = stop - start
-                trimmed.append(basket_arrays[basket_num][local_start:local_stop])
+                to_append = basket_arrays[basket_num][local_start:local_stop]
 
             elif start <= entry_stop <= stop:
                 local_start = 0
                 local_stop = entry_stop - start
-                trimmed.append(basket_arrays[basket_num][local_start:local_stop])
+                to_append = basket_arrays[basket_num][local_start:local_stop]
 
             elif entry_start < stop and start <= entry_stop:
-                trimmed.append(basket_arrays[basket_num])
+                to_append = basket_arrays[basket_num]
+
+            if to_append is not None and has_any_awkward_types:
+
+                if isinstance(library, uproot.interpretation.library.NumPy):
+                    trimmed.append(to_append)
+
+                elif isinstance(library, uproot.interpretation.library.Awkward):
+
+                    if isinstance(to_append, numpy.ndarray):
+                        trimmed.append(
+                            uproot.interpretation.library._object_to_awkward_array(
+                                uproot.extras.awkward(), self._form, to_append
+                            )
+                        )
+                    else:
+                        trimmed.append(to_append)
+
+                elif isinstance(library, uproot.interpretation.library.Pandas):
+
+                    if isinstance(to_append, numpy.ndarray):
+                        trimmed.append(
+                            uproot.interpretation.library._process_array_for_pandas(
+                                to_append,
+                                False,
+                                branch.file.interpretation,
+                                form=self._form,
+                            )
+                        )
+                    else:
+                        trimmed.append(to_append)
+
+            elif to_append is not None:
+                trimmed.append(to_append)
 
             start = stop
 
-        if all(uproot._util.from_module(x, "awkward") for x in basket_arrays.values()):
-            assert isinstance(library, uproot.interpretation.library.Awkward)
-            awkward = library.imported
+        if len(basket_arrays) == 0:
+            output = numpy.array([], dtype=self.numpy_dtype)
+
+        elif all(
+            uproot._util.from_module(x, "awkward") for x in basket_arrays.values()
+        ) and isinstance(
+            library,
+            (
+                uproot.interpretation.library.Awkward,
+                uproot.interpretation.library.Pandas,
+            ),
+        ):
+            awkward = uproot.extras.awkward()
             output = awkward.concatenate(trimmed, mergebool=False, highlevel=False)
         else:
             output = numpy.concatenate(trimmed)
@@ -222,7 +522,9 @@ class AsObjects(uproot.interpretation.Interpretation):
             output=output,
         )
 
-        output = library.finalize(output, branch, self, entry_start, entry_stop)
+        output = library.finalize(
+            output, branch, self, entry_start, entry_stop, options
+        )
 
         self.hook_after_final_array(
             basket_arrays=basket_arrays,
@@ -232,6 +534,7 @@ class AsObjects(uproot.interpretation.Interpretation):
             library=library,
             branch=branch,
             output=output,
+            options=options,
         )
 
         return output
@@ -245,7 +548,7 @@ class AsObjects(uproot.interpretation.Interpretation):
         ``self``.
         """
         if self._branch is not None:
-            try:
+            with contextlib.suppress(CannotBeStrided):
                 return self._model.strided_interpretation(
                     self._branch.file,
                     header=False,
@@ -253,8 +556,6 @@ class AsObjects(uproot.interpretation.Interpretation):
                     breadcrumbs=(),
                     original=self._model,
                 )
-            except CannotBeStrided:
-                pass
 
         if isinstance(self._model, uproot.containers.AsString):
             header_bytes = 0
@@ -296,7 +597,7 @@ class AsObjects(uproot.interpretation.Interpretation):
                 )
 
             if self._branch is not None:
-                try:
+                with contextlib.suppress(CannotBeStrided):
                     content = self._model.values.strided_interpretation(
                         self._branch.file,
                         header=False,
@@ -319,8 +620,6 @@ class AsObjects(uproot.interpretation.Interpretation):
                         self._model.typename,
                         original=self._model,
                     )
-                except CannotBeStrided:
-                    pass
 
         return self
 
@@ -339,14 +638,31 @@ def _unravel_members(members):
 def _strided_awkward_form(awkward, classname, members, file, context):
     contents = {}
     for name, member in members:
-        if isinstance(member, AsStridedObjects):
-            cname = uproot.model.classname_decode(member._model.__name__)[0]
-            contents[name] = _strided_awkward_form(
-                awkward, cname, member._members, file, context
-            )
+        if not context["header"] and name in ("@num_bytes", "@instance_version"):
+            pass
+        elif not context["tobject_header"] and name in (
+            None,
+            "@num_bytes",
+            "@instance_version",
+            "@fUniqueID",
+            "@fBits",
+            "@pidf",
+        ):
+            pass
         else:
-            contents[name] = uproot._util.awkward_form(member, file, context)
-    return awkward.forms.RecordForm(contents, parameters={"__record__": classname})
+            if isinstance(member, AsStridedObjects):
+                cname = uproot.model.classname_decode(member._model.__name__)[0]
+                contents[name] = _strided_awkward_form(
+                    awkward, cname, member._members, file, context
+                )
+            else:
+                contents[name] = uproot._util.awkward_form(member, file, context)
+
+    return awkward.forms.RecordForm(
+        list(contents.values()),
+        list(contents.keys()),
+        parameters={"__record__": classname},
+    )
 
 
 class AsStridedObjects(uproot.interpretation.numerical.AsDtype):
@@ -381,10 +697,25 @@ class AsStridedObjects(uproot.interpretation.numerical.AsDtype):
     """
 
     def __init__(self, model, members, original=None):
+        all_headers_prepended = False
+
+        for first_value_loc in range(len(members)):
+            if members[first_value_loc] != (None, None):
+                break
+
+        for i in range(first_value_loc, len(members)):
+            member, value = members[i]
+            if member is not None and not all_headers_prepended:
+                all_headers_prepended = True
+            if member is None and all_headers_prepended or len(members) == 1:
+                all_headers_prepended = False
+                del members[i]
+
         self._model = model
-        self._members = members
+        self._members = members[first_value_loc:]
         self._original = original
-        super().__init__(_unravel_members(members))
+        self._all_headers_prepended = all_headers_prepended
+        super().__init__(_unravel_members(self._members))
 
     @property
     def model(self):
@@ -415,9 +746,7 @@ class AsStridedObjects(uproot.interpretation.numerical.AsDtype):
 
     def __repr__(self):
         if self.inner_shape:
-            return "AsStridedObjects({}, {})".format(
-                self._model.__name__, self.inner_shape
-            )
+            return f"AsStridedObjects({self._model.__name__}, {self.inner_shape})"
         else:
             return f"AsStridedObjects({self._model.__name__})"
 
@@ -428,13 +757,17 @@ class AsStridedObjects(uproot.interpretation.numerical.AsDtype):
     def numpy_dtype(self):
         return numpy.dtype(object)
 
+    @property
+    def all_headers_prepended(self):
+        return self._all_headers_prepended
+
     def awkward_form(
         self,
         file,
         context=None,
         index_format="i64",
         header=False,
-        tobject_header=True,
+        tobject_header=False,
         breadcrumbs=(),
     ):
         context = self._make_context(
@@ -458,6 +791,64 @@ class AsStridedObjects(uproot.interpretation.numerical.AsDtype):
     def _wrap_almost_finalized(self, array):
         return StridedObjectArray(self, array)
 
+    def basket_array(
+        self,
+        data,
+        byte_offsets,
+        basket,
+        branch,
+        context,
+        cursor_offset,
+        library,
+        options,
+    ):
+        self.hook_before_basket_array(
+            data=data,
+            byte_offsets=byte_offsets,
+            basket=basket,
+            branch=branch,
+            context=context,
+            cursor_offset=cursor_offset,
+            library=library,
+            options=options,
+        )
+
+        dtype, shape = uproot.interpretation.numerical._dtype_shape(self._from_dtype)
+
+        if (
+            byte_offsets is not None
+            and dtype.itemsize != byte_offsets[1] - byte_offsets[0]
+            and self.all_headers_prepended
+        ):
+            dtype = [
+                ("@headers", "u1", byte_offsets[1] - byte_offsets[0] - dtype.itemsize)
+            ] + [
+                (x, str(y[0]))
+                for x, y in sorted(dtype.fields.items(), key=lambda k: k[1])
+            ]
+            self._to_dtype = numpy.dtype(dtype)
+        try:
+            output = data.view(dtype).reshape((-1, *shape))
+
+        except ValueError as err:
+            raise ValueError(
+                f"""basket {basket.basket_num} in tree/branch {branch.object_path} has the wrong number of bytes ({len(data)}) """
+                f"""for interpretation {self}
+in file {branch.file.file_path}"""
+            ) from err
+        self.hook_after_basket_array(
+            data=data,
+            byte_offsets=byte_offsets,
+            basket=basket,
+            branch=branch,
+            context=context,
+            output=output,
+            cursor_offset=cursor_offset,
+            library=library,
+            options=options,
+        )
+        return output
+
 
 class CannotBeStrided(Exception):
     """
@@ -467,7 +858,13 @@ class CannotBeStrided(Exception):
     non-conforming type is found.
     """
 
-    pass
+
+class CannotBeForth(Exception):
+    """
+    Exception used to stop recursion over
+    :ref:`uproot.model.Model.read` as soon as a
+    non-conforming type is found.
+    """
 
 
 class CannotBeAwkward(Exception):
@@ -522,14 +919,7 @@ class ObjectArray:
         self._detached_file = self._branch.file.detached
 
     def __repr__(self):
-        return "ObjectArray({}, {}, {}, {}, {}, {})".format(
-            self._model,
-            self._branch,
-            self._context,
-            self._byte_offsets,
-            self._byte_content,
-            self._cursor_offset,
-        )
+        return f"ObjectArray({self._model}, {self._branch}, {self._context}, {self._byte_offsets}, {self._byte_content}, {self._cursor_offset})"
 
     @property
     def model(self):
@@ -686,7 +1076,7 @@ class StridedObjectArray:
 
     @property
     def shape(self):
-        assert self._array.shape == (len(self),) + self._interpretation.inner_shape
+        assert self._array.shape == (len(self), *self._interpretation.inner_shape)
         return self._array.shape
 
     def __repr__(self):
