@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines languages for expressions passed to
@@ -8,6 +8,7 @@ The default is :doc:`uproot.language.python.PythonLanguage`.
 
 All languages must be subclasses of :doc:`uproot.language.Language`.
 """
+from __future__ import annotations
 
 
 class Language:
@@ -18,5 +19,3 @@ class Language:
 
     The default is :doc:`uproot.language.python.PythonLanguage`.
     """
-
-    pass
