@@ -1,9 +1,9 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines versionless models for ``TArray`` and its subclasses.
 """
-
+from __future__ import annotations
 
 import struct
 from collections.abc import Sequence
@@ -26,12 +26,12 @@ class Model_TArray(uproot.model.Model, Sequence):
         pass
 
     def read_members(self, chunk, cursor, context, file):
+        if uproot._awkwardforth.get_forth_obj(context) is not None:
+            raise uproot.interpretation.objects.CannotBeForth()
         if self.is_memberwise:
             raise NotImplementedError(
-                """memberwise serialization of {}
-in file {}""".format(
-                    type(self).__name__, self.file.file_path
-                )
+                f"""memberwise serialization of {type(self).__name__}
+in file {self.file.file_path}"""
             )
         self._members["fN"] = cursor.field(chunk, _tarray_format1, context)
         self._data = cursor.array(chunk, self._members["fN"], self.dtype, context)
@@ -79,7 +79,6 @@ in file {}""".format(
         return awkward.forms.ListOffsetForm(
             context["index_format"],
             uproot._util.awkward_form(cls.dtype, file, context),
-            parameters={"uproot": {"as": "TArray"}},
         )
 
     writable = True
