@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines the entry-point for opening a file, :doc:`uproot.reading.open`,
@@ -7,27 +7,31 @@ and the classes that are too fundamental to be models:
 :doc:`uproot.reading.ReadOnlyDirectory` (``TDirectory`` or ``TDirectoryFile``),
 and :doc:`uproot.reading.ReadOnlyKey` (``TKey``).
 """
+from __future__ import annotations
 
-
+import re
 import struct
 import sys
 import uuid
-import warnings
 from collections.abc import Mapping, MutableMapping
+from pathlib import Path
+from typing import IO
 
 import uproot
 import uproot.behaviors.TBranch
+import uproot.source.fsspec
 from uproot._util import no_filter
 
 
 def open(
-    path,
+    path: str | Path | IO | dict[str | Path | IO, str],
+    *,
     object_cache=100,
     array_cache="100 MB",
     custom_classes=None,
     decompression_executor=None,
     interpretation_executor=None,
-    **options,  # NOTE: a comma after **options breaks Python 2
+    **options,
 ):
     """
     Args:
@@ -42,7 +46,7 @@ def open(
             ``"rel/file.root:tdirectory/ttree"``, ``Path("rel:/file.root")``,
             ``Path("/abs/path:stuff.root")``
         object_cache (None, MutableMapping, or int): Cache of objects drawn
-            from ROOT directories (e.g histograms, TTrees, other directories);
+            from ROOT directories (e.g. histograms, TTrees, other directories);
             if None, do not use a cache; if an int, create a new cache of this
             size.
         array_cache (None, MutableMapping, or memory size): Cache of arrays
@@ -76,13 +80,11 @@ def open(
 
     Options (type; default):
 
-    * file_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.file.MemmapSource`)
-    * xrootd_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.xrootd.XRootDSource`)
-    * http_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.http.HTTPSource`)
-    * object_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.object.ObjectSource`)
+    * handler (:doc:`uproot.source.chunk.Source` class; None)
     * timeout (float for HTTP, int for XRootD; 30)
     * max_num_elements (None or int; None)
     * num_workers (int; 1)
+    * use_threads (bool; False on the emscripten platform (i.e. in a web browser), else True)
     * num_fallback_workers (int; 10)
     * begin_chunk_size (memory_size; 403, the smallest a ROOT file can be)
     * minimal_ttree_metadata (bool; True)
@@ -108,8 +110,7 @@ def open(
       contiguous entries in ``TTrees``.
     * :doc:`uproot.behaviors.TBranch.concatenate`: returns a single concatenated
       array from ``TTrees``.
-    * :doc:`uproot.behaviors.TBranch.lazy`: returns a lazily read array from
-      ``TTrees``.
+    * :doc:`uproot._dask.dask`: returns an unevaluated Dask array from ``TTrees``.
 
     For remote ROOT files served over HTTP(S), basic authentication is supported.
     In this case, the credentials may be provided part of the URL in, as in
@@ -120,7 +121,7 @@ def open(
     if isinstance(path, dict) and len(path) == 1:
         ((file_path, object_path),) = path.items()
 
-    elif uproot._util.isstr(path):
+    elif isinstance(path, str):
         file_path, object_path = uproot._util.file_object_path_split(path)
 
     else:
@@ -129,13 +130,13 @@ def open(
 
     file_path = uproot._util.regularize_path(file_path)
 
-    if not uproot._util.isstr(file_path) and not (
+    if not isinstance(file_path, str) and not (
         hasattr(file_path, "read") and hasattr(file_path, "seek")
     ):
         raise ValueError(
             "'path' must be a string, pathlib.Path, an object with 'read' and "
-            "'seek' methods, or a length-1 dict of {{file_path: object_path}}, "
-            "not {0}".format(repr(path))
+            "'seek' methods, or a length-1 dict of {file_path: object_path}, "
+            f"not {path!r}"
         )
 
     file = ReadOnlyFile(
@@ -145,7 +146,7 @@ def open(
         custom_classes=custom_classes,
         decompression_executor=decompression_executor,
         interpretation_executor=interpretation_executor,
-        **options,  # NOTE: a comma after **options breaks Python 2
+        **options,
     )
 
     if object_path is None:
@@ -154,41 +155,22 @@ def open(
         return file.root_directory[object_path]
 
 
+open.defaults = {
+    "handler": None,
+    "timeout": 30,
+    "max_num_elements": None,
+    "num_workers": 1,
+    "use_threads": sys.platform != "emscripten",
+    "num_fallback_workers": 10,
+    "begin_chunk_size": 403,  # the smallest a ROOT file can be
+    "minimal_ttree_metadata": True,
+    "http_max_header_bytes": 21784,
+}
+
+
 class _OpenDefaults(dict):
-    def __getitem__(self, where):
-        if where == "xrootd_handler" and where not in self:
-            # See https://github.com/scikit-hep/uproot4/issues/294
-            if uproot.extras.older_xrootd("5.2.0"):
-                message = (
-                    "XRootD {} is not fully supported; ".format(
-                        uproot.extras.xrootd_version()
-                    )
-                    + """either upgrade to 5.2.0+ or set
-
-    open.defaults["xrootd_handler"] = uproot.MultithreadedXRootDSource
-"""
-                )
-                warnings.warn(message, FutureWarning)
-
-            # The key should still be set, regardless of whether we see the warning.
-            self["xrootd_handler"] = uproot.source.xrootd.XRootDSource
-
-        return dict.__getitem__(self, where)
-
-
-open.defaults = _OpenDefaults(
-    {
-        "file_handler": uproot.source.file.MemmapSource,
-        "http_handler": uproot.source.http.HTTPSource,
-        "object_handler": uproot.source.object.ObjectSource,
-        "timeout": 30,
-        "max_num_elements": None,
-        "num_workers": 1,
-        "num_fallback_workers": 10,
-        "begin_chunk_size": 403,  # the smallest a ROOT file can be
-        "minimal_ttree_metadata": True,
-    }
-)
+    def __init__(self):
+        raise NotImplementedError  # kept for backwards compatibility
 
 
 must_be_attached = [
@@ -532,13 +514,11 @@ class ReadOnlyFile(CommonFileMethods):
 
     Options (type; default):
 
-    * file_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.file.MemmapSource`)
-    * xrootd_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.xrootd.XRootDSource`)
-    * http_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.http.HTTPSource`)
-    * object_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.object.ObjectSource`)
+    * handler (:doc:`uproot.source.chunk.Source` class; None)
     * timeout (float for HTTP, int for XRootD; 30)
     * max_num_elements (None or int; None)
     * num_workers (int; 1)
+    * use_threads (bool; False on the emscripten platform (i.e. in a web browser), else True)
     * num_fallback_workers (int; 10)
     * begin_chunk_size (memory_size; 403, the smallest a ROOT file can be)
     * minimal_ttree_metadata (bool; True)
@@ -549,13 +529,14 @@ class ReadOnlyFile(CommonFileMethods):
 
     def __init__(
         self,
-        file_path,
+        file_path: str | Path | IO,
+        *,
         object_cache=100,
         array_cache="100 MB",
         custom_classes=None,
         decompression_executor=None,
         interpretation_executor=None,
-        **options,  # NOTE: a comma after **options breaks Python 2
+        **options,
     ):
         self._file_path = file_path
         self.object_cache = object_cache
@@ -564,7 +545,7 @@ class ReadOnlyFile(CommonFileMethods):
         self.decompression_executor = decompression_executor
         self.interpretation_executor = interpretation_executor
 
-        self._options = _OpenDefaults(open.defaults)
+        self._options = open.defaults.copy()
         self._options.update(options)
         for option in ["begin_chunk_size"]:
             self._options[option] = uproot._util.memory_size(self._options[option])
@@ -574,12 +555,10 @@ class ReadOnlyFile(CommonFileMethods):
 
         self.hook_before_create_source()
 
-        Source, file_path = uproot._util.file_path_to_source_class(
+        source_cls, file_path = uproot._util.file_path_to_source_class(
             file_path, self._options
         )
-        self._source = Source(
-            file_path, **self._options  # NOTE: a comma after **options breaks Python 2
-        )
+        self._source = source_cls(file_path, **self._options)
 
         self.hook_before_get_chunks()
 
@@ -640,10 +619,8 @@ class ReadOnlyFile(CommonFileMethods):
 
         if magic != b"root":
             raise ValueError(
-                """not a ROOT file: first four bytes are {}
-in file {}""".format(
-                    repr(magic), file_path
-                )
+                f"""not a ROOT file: first four bytes are {magic!r}
+in file {file_path}"""
             )
 
     def __repr__(self):
@@ -768,7 +745,7 @@ in file {}""".format(
     def array_cache(self, value):
         if value is None or isinstance(value, MutableMapping):
             self._array_cache = value
-        elif uproot._util.isint(value) or uproot._util.isstr(value):
+        elif uproot._util.isint(value) or isinstance(value, str):
             self._array_cache = uproot.cache.LRUArrayCache(value)
         else:
             raise TypeError(
@@ -918,10 +895,8 @@ in file {}""".format(
 
                     else:
                         raise ValueError(
-                            """unexpected type in TList of streamers and streamer rules: {}
-in file {}""".format(
-                                type(x), self._file_path
-                            )
+                            f"""unexpected type in TList of streamers and streamer rules: {type(x)}
+in file {self._file_path}"""
                         )
 
                 self.hook_after_interpret_streamers(
@@ -948,7 +923,7 @@ in file {}""".format(
         :ref:`uproot.reading.ReadOnlyFile.streamers`.
         """
         if self._streamer_rules is None:
-            self.streamers
+            self.streamers  # noqa: B018 (this is not a useless expression; it runs code)
         return self._streamer_rules
 
     def streamers_named(self, classname):
@@ -1131,6 +1106,16 @@ in file {}""".format(
             if len(streamers) == 0 and self._custom_classes is not None:
                 cls = uproot.classes.get(classname)
 
+        if (
+            cls is None
+            and re.match(r"(std\s*::\s*)?(vector|list|map|set|bitset)\s*<", classname)
+            is not None
+        ):
+            cls = uproot.interpretation.identify.parse_typename(classname)
+            cls._header = False
+
+            return cls
+
         if cls is None:
             if len(streamers) == 0:
                 unknown_cls = uproot.unknown_classes.get(classname)
@@ -1216,7 +1201,6 @@ in file {}""".format(
         This is the first hook called in the :doc:`uproot.reading.ReadOnlyFile`
         constructor.
         """
-        pass
 
     def hook_before_get_chunks(self, **kwargs):
         """
@@ -1225,7 +1209,6 @@ in file {}""".format(
         get any :doc:`uproot.source.chunk.Chunk`, specifically the
         :ref:`uproot.reading.ReadOnlyFile.begin_chunk`.
         """
-        pass
 
     def hook_before_interpret(self, **kwargs):
         """
@@ -1233,7 +1216,6 @@ in file {}""".format(
         loading the :ref:`uproot.reading.ReadOnlyFile.begin_chunk` and before
         interpreting its ``TFile`` header.
         """
-        pass
 
     def hook_after_interpret(self, **kwargs):
         """
@@ -1244,7 +1226,6 @@ in file {}""".format(
         This is the last hook called in the :doc:`uproot.reading.ReadOnlyFile`
         constructor.
         """
-        pass
 
     def hook_before_read_streamer_key(self, **kwargs):
         """
@@ -1254,7 +1235,6 @@ in file {}""".format(
         This is the first hook called in
         :ref:`uproot.reading.ReadOnlyFile.streamers`.
         """
-        pass
 
     def hook_before_read_decompress_streamers(self, **kwargs):
         """
@@ -1262,7 +1242,6 @@ in file {}""".format(
         the ``TKey`` associated with the ``TStreamerInfo`` and before reading
         and decompressing the ``TStreamerInfo`` data.
         """
-        pass
 
     def hook_before_interpret_streamers(self, **kwargs):
         """
@@ -1270,7 +1249,6 @@ in file {}""".format(
         and decompressing the ``TStreamerInfo`` data, but before interpreting
         it.
         """
-        pass
 
     def hook_after_interpret_streamers(self, **kwargs):
         """
@@ -1280,7 +1258,6 @@ in file {}""".format(
         This is the last hook called in
         :ref:`uproot.reading.ReadOnlyFile.streamers`.
         """
-        pass
 
 
 _directory_format_small = struct.Struct(">hIIiiiii")
@@ -1491,7 +1468,7 @@ class ReadOnlyDirectory(Mapping):
         See :ref:`uproot.reading.ReadOnlyDirectory.path` for the path as a
         tuple of strings.
         """
-        return "/".join(("",) + self._path + ("",)).replace("//", "/")
+        return "/".join(("", *self._path, "")).replace("//", "/")
 
     @property
     def file_path(self):
@@ -1599,6 +1576,7 @@ class ReadOnlyDirectory(Mapping):
 
     def keys(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -1631,6 +1609,7 @@ class ReadOnlyDirectory(Mapping):
 
     def values(
         self,
+        *,
         recursive=True,
         filter_name=no_filter,
         filter_classname=no_filter,
@@ -1661,6 +1640,7 @@ class ReadOnlyDirectory(Mapping):
 
     def items(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -1694,6 +1674,7 @@ class ReadOnlyDirectory(Mapping):
 
     def classnames(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -1726,6 +1707,7 @@ class ReadOnlyDirectory(Mapping):
 
     def iterkeys(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -1775,6 +1757,7 @@ class ReadOnlyDirectory(Mapping):
 
     def itervalues(
         self,
+        *,
         recursive=True,
         filter_name=no_filter,
         filter_classname=no_filter,
@@ -1805,6 +1788,7 @@ class ReadOnlyDirectory(Mapping):
 
     def iteritems(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -1855,6 +1839,7 @@ class ReadOnlyDirectory(Mapping):
 
     def iterclassnames(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -1928,7 +1913,24 @@ class ReadOnlyDirectory(Mapping):
     def __iter__(self):
         return self.iterkeys()
 
-    def title_of(self, where):
+    def descent_into_path(self, where):
+        items = re.split("[:/]", where)
+        step = last = self
+        for item in items[:-1]:
+            if item != "":
+                if isinstance(step, ReadOnlyDirectory):
+                    last = step
+                    step = step[item]
+                else:
+                    raise uproot.KeyInFileError(
+                        where,
+                        because=repr(item) + " is not a TDirectory",
+                        keys=[key.fName for key in last._keys],
+                        file_path=self._file.file_path,
+                    )
+        return step, items[-1]
+
+    def title_of(self, where, recursive=True):
         """
         Returns the title of the object selected by ``where``.
 
@@ -1941,9 +1943,13 @@ class ReadOnlyDirectory(Mapping):
 
         Note that this does not read any data from the file.
         """
-        return self.key(where).title()
+        if recursive and "/" in where or ":" in where:
+            step, last_item = self.descent_into_path(where)
+            return step[last_item].title
+        else:
+            return self.key(where).title()
 
-    def classname_of(self, where, encoded=False, version=None):
+    def classname_of(self, where, encoded=False, version=None, recursive=True):
         """
         Returns the classname of the object selected by ``where``. If
         ``encoded`` with a possible ``version``, return a Python classname;
@@ -1958,10 +1964,14 @@ class ReadOnlyDirectory(Mapping):
 
         Note that this does not read any data from the file.
         """
-        key = self.key(where)
-        return key.classname(encoded=encoded, version=version)
 
-    def class_of(self, where, version=None):
+        if recursive and "/" in where or ":" in where:
+            step, last_item = self.descent_into_path(where)
+            return step[last_item].classname
+        else:
+            return self.key(where).classname(encoded=encoded, version=version)
+
+    def class_of(self, where, version=None, recursive=True):
         """
         Returns a class object for the ROOT object selected by ``where``. If
         ``version`` is specified, get a :doc:`uproot.model.VersionedModel`;
@@ -1977,10 +1987,15 @@ class ReadOnlyDirectory(Mapping):
 
         Note that this does not read any data from the file.
         """
-        key = self.key(where)
-        return self._file.class_named(key.fClassName, version=version)
+        if recursive and "/" in where or ":" in where:
+            return self._file.class_named(
+                self.classname_of(where, version=version), version=version
+            )
+        else:
+            key = self.key(where)
+            return self._file.class_named(key.fClassName, version=version)
 
-    def streamer_of(self, where, version="max"):
+    def streamer_of(self, where, version="max", recursive=True):
         """
         Returns a ``TStreamerInfo`` (:doc:`uproot.streamers.Model_TStreamerInfo`)
         for the object selected by ``where`` and ``version``.
@@ -1994,8 +2009,13 @@ class ReadOnlyDirectory(Mapping):
 
         Note that this does not read any data from the file.
         """
-        key = self.key(where)
-        return self._file.streamer_named(key.fClassName, version)
+        if recursive and "/" in where or ":" in where:
+            return self._file.streamer_named(
+                self.classname_of(where, version=version), version=version
+            )
+        else:
+            key = self.key(where)
+            return self._file.streamer_named(key.fClassName, version=version)
 
     def key(self, where):
         """
@@ -2014,21 +2034,8 @@ class ReadOnlyDirectory(Mapping):
         where = uproot._util.ensure_str(where)
 
         if "/" in where:
-            items = where.split("/")
-            step = last = self
-            for item in items[:-1]:
-                if item != "":
-                    if isinstance(step, ReadOnlyDirectory):
-                        last = step
-                        step = step[item]
-                    else:
-                        raise uproot.KeyInFileError(
-                            where,
-                            because=repr(item) + " is not a TDirectory",
-                            keys=[key.fName for key in last._keys],
-                            file_path=self._file.file_path,
-                        )
-            return step.key(items[-1])
+            step, last_item = self.descent_into_path(where)
+            return step.key(last_item)
 
         if ";" in where:
             at = where.rindex(";")
@@ -2047,7 +2054,8 @@ class ReadOnlyDirectory(Mapping):
                 return key
             elif cycle is None and last is None:
                 last = key
-            elif cycle is None and last.fCycle < key.fCycle:
+            # Follow ROOT's behaviour in comparing negative fCycle values
+            elif cycle is None and abs(last.fCycle) < abs(key.fCycle):
                 last = key
 
         if last is not None:
@@ -2084,6 +2092,8 @@ class ReadOnlyDirectory(Mapping):
                                     keys=[key.fName for key in last._keys],
                                     file_path=self._file.file_path,
                                 )
+                        elif ":" in item and item in step:
+                            return self.key(where).get()
                         else:
                             last = step
                             step = step[item]
@@ -2177,14 +2187,12 @@ class ReadOnlyDirectory(Mapping):
         This is the first hook called in the
         :doc:`uproot.reading.ReadOnlyDirectory` constructor.
         """
-        pass
 
     def hook_before_interpret(self, **kwargs):
         """
         Called in the :doc:`uproot.reading.ReadOnlyDirectory` constructor after
         reading the ``TDirectory`` header fields and before interpreting them.
         """
-        pass
 
     def hook_before_read_keys(self, **kwargs):
         """
@@ -2192,14 +2200,12 @@ class ReadOnlyDirectory(Mapping):
         interpreting the ``TDirectory`` header fields and before reading the
         chunk of ``TKeys``.
         """
-        pass
 
     def hook_before_header_key(self, **kwargs):
         """
         Called in the :doc:`uproot.reading.ReadOnlyDirectory` constructor after
         reading the chunk of ``TKeys`` and before interpreting the header ``TKey``.
         """
-        pass
 
     def hook_before_keys(self, **kwargs):
         """
@@ -2207,7 +2213,6 @@ class ReadOnlyDirectory(Mapping):
         interpreting the header ``TKey`` and number of keys, and before
         interpeting the object ``TKeys``.
         """
-        pass
 
     def hook_after_keys(self, **kwargs):
         """
@@ -2217,7 +2222,6 @@ class ReadOnlyDirectory(Mapping):
         This is the last hook called in the
         :doc:`uproot.reading.ReadOnlyDirectory` constructor.
         """
-        pass
 
 
 _key_format_small = struct.Struct(">ihiIhhii")
@@ -2317,9 +2321,7 @@ class ReadOnlyKey:
             nameclass = ""
         else:
             nameclass = f" {self.name(cycle=True)}: {self.classname()}"
-        return "<ReadOnlyKey{} (seek pos {}) at 0x{:012x}>".format(
-            nameclass, self.data_cursor.index, id(self)
-        )
+        return f"<ReadOnlyKey{nameclass} (seek pos {self.data_cursor.index}) at 0x{id(self):012x}>"
 
     @property
     def cursor(self):
@@ -2392,9 +2394,7 @@ class ReadOnlyKey:
         ``TDirectory``), this returns a message with the raw seek position.
         """
         if isinstance(self._parent, ReadOnlyDirectory):
-            return "{}{};{}".format(
-                self._parent.object_path, self.name(False), self._fCycle
-            )
+            return f"{self._parent.object_path}{self.name(False)};{self._fCycle}"
         else:
             return f"(seek pos {self.data_cursor.index})/{self.name(False)}"
 
@@ -2492,7 +2492,7 @@ class ReadOnlyKey:
             "TDirectoryFile",
         ):
             out = ReadOnlyDirectory(
-                self._parent.path + (self.fName,),
+                (*self._parent.path, self.fName),
                 self.data_cursor,
                 {},
                 self._file,
@@ -2502,9 +2502,13 @@ class ReadOnlyKey:
         else:
             chunk, cursor = self.get_uncompressed_chunk_cursor()
             start_cursor = cursor.copy()
-            cls = self._file.class_named(self._fClassName)
             context = {"breadcrumbs": (), "TKey": self}
 
+            if re.match(r"(std\s*::\s*)?string", self._fClassName):
+                return cursor.string(chunk, context)
+
+            cls = self._file.class_named(self._fClassName)
+
             try:
                 out = cls.read(chunk, cursor, context, self._file, selffile, parent)
 
@@ -2676,7 +2680,6 @@ class ReadOnlyKey:
         This is the first hook called in the
         :doc:`uproot.reading.ReadOnlyKey` constructor.
         """
-        pass
 
     def hook_before_strings(self, **kwargs):
         """
@@ -2688,7 +2691,6 @@ class ReadOnlyKey:
 
         Only called if ``read_strings=True`` is passed to the constructor.
         """
-        pass
 
     def hook_after_interpret(self, **kwargs):
         """
@@ -2698,4 +2700,3 @@ class ReadOnlyKey:
         This is the last hook called in the
         :doc:`uproot.reading.ReadOnlyKey` constructor.
         """
-        pass
