@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines a Python-like Future and Executor for Uproot in three levels:
@@ -21,12 +21,14 @@ This module defines a Python-like Future and Executor for Uproot in three levels
 These classes implement a *subset* of Python's Future and Executor interfaces.
 """
 
+from __future__ import annotations
 
 import os
 import queue
 import sys
 import threading
 import time
+from abc import ABC, abstractmethod
 
 
 def delayed_raise(exception_class, exception_value, traceback):
@@ -36,6 +38,32 @@ def delayed_raise(exception_class, exception_value, traceback):
     raise exception_value.with_traceback(traceback)
 
 
+class Executor(ABC):
+    def __repr__(self):
+        return f"<{self.__class__.__name__} at 0x{id(self):012x}>"
+
+    @abstractmethod
+    def submit(self, task, /, *args, **kwargs):
+        """
+        Submit a task to be run in the background and return a Future object
+        representing that task.
+        """
+        raise NotImplementedError
+
+    def shutdown(self, wait: bool = True):
+        """
+        Stop the executor and free its resources.
+        """
+        return
+
+    @property
+    def closed(self) -> bool:
+        """
+        True if the executor has been stopped and its resources freed.
+        """
+        return False
+
+
 ##################### use-case 1: trivial Futures/Executor (satisfying formalities)
 
 
@@ -48,6 +76,12 @@ class TrivialFuture:
     def __init__(self, result):
         self._result = result
 
+    def add_done_callback(self, callback, *, context=None):
+        """
+        The callback is called immediately.
+        """
+        callback(self)
+
     def result(self, timeout=None):
         """
         The result of this (Trivial)Future.
@@ -55,7 +89,7 @@ class TrivialFuture:
         return self._result
 
 
-class TrivialExecutor:
+class TrivialExecutor(Executor):
     """
     Formally satisfies the interface for a
     :doc:`uproot.source.futures.ThreadPoolExecutor`, but the
@@ -63,20 +97,11 @@ class TrivialExecutor:
     ``task`` synchronously.
     """
 
-    def __repr__(self):
-        return f"<TrivialExecutor at 0x{id(self):012x}>"
-
-    def submit(self, task, *args):
+    def submit(self, task, /, *args, **kwargs):
         """
         Immediately runs ``task(*args)``.
         """
-        return TrivialFuture(task(*args))
-
-    def shutdown(self, wait=True):
-        """
-        Does nothing, since this object does not have threads to stop.
-        """
-        pass
+        return TrivialFuture(task(*args, **kwargs))
 
 
 ##################### use-case 2: Python-like Futures/Executor for compute
@@ -140,13 +165,13 @@ class Worker(threading.Thread):
     :doc:`uproot.source.futures.ThreadPoolExecutor`.
     """
 
-    def __init__(self, work_queue):
+    def __init__(self, work_queue: queue.Queue):
         super().__init__()
         self.daemon = True
         self._work_queue = work_queue
 
     @property
-    def work_queue(self):
+    def work_queue(self) -> queue.Queue:
         """
         The worker calls ``get`` on this queue for tasks in the form of
         :doc:`uproot.source.futures.Future` objects and runs them. If it ever
@@ -170,11 +195,12 @@ class Worker(threading.Thread):
             future._run()
 
 
-class ThreadPoolExecutor:
+class ThreadPoolExecutor(Executor):
     """
     Args:
-        num_workers (None or int): The number of workers to start. If None,
-            use ``os.cpu_count()``.
+        max_workers (None or int): The maximum number of workers to start.
+        In the current implementation this is exactly the number of workers.
+        If None, use ``os.cpu_count()``.
 
     Like Python 3 ``concurrent.futures.ThreadPoolExecutor`` except that it has
     only the subset of the interface Uproot needs and is available in Python 2.
@@ -183,42 +209,47 @@ class ThreadPoolExecutor:
     class.
     """
 
-    def __init__(self, num_workers=None):
-        if num_workers is None:
+    def __init__(self, max_workers: int | None = None):
+        if max_workers is None:
             if hasattr(os, "cpu_count"):
-                num_workers = os.cpu_count()
+                self._max_workers = os.cpu_count()
             else:
                 import multiprocessing
 
-                num_workers = multiprocessing.cpu_count()
+                self._max_workers = multiprocessing.cpu_count()
 
         self._work_queue = queue.Queue()
         self._workers = []
-        for _ in range(num_workers):
+        for _ in range(self._max_workers):
             self._workers.append(Worker(self._work_queue))
         for worker in self._workers:
             worker.start()
 
     def __repr__(self):
-        return "<ThreadPoolExecutor ({} workers) at 0x{:012x}>".format(
-            len(self._workers), id(self)
-        )
+        return f"<{self.__class__.__name__} ({len(self._workers)} workers) at 0x{id(self):012x}>"
 
     @property
-    def num_workers(self):
+    def max_workers(self) -> int:
+        """
+        The maximum number of workers.
+        """
+        return self._max_workers
+
+    @property
+    def num_workers(self) -> int:
         """
         The number of workers.
         """
         return len(self._workers)
 
     @property
-    def workers(self):
+    def workers(self) -> list[Worker]:
         """
         A list of workers (:doc:`uproot.source.futures.Worker`).
         """
         return self._workers
 
-    def submit(self, task, *args):
+    def submit(self, task, /, *args, **kwargs):
         """
         Pass the ``task`` and ``args`` onto the workers'
         :ref:`uproot.source.futures.Worker.work_queue` as a
@@ -229,7 +260,7 @@ class ThreadPoolExecutor:
         self._work_queue.put(future)
         return future
 
-    def shutdown(self, wait=True):
+    def shutdown(self, wait: bool = True):
         """
         Stop every :doc:`uproot.source.futures.Worker` by putting None
         on the :ref:`uproot.source.futures.Worker.work_queue` until none of
@@ -264,6 +295,9 @@ class ResourceFuture(Future):
         self._notify = None
 
     def _set_notify(self, notify):
+        """
+        Set the ``notify`` function that is called when this task is complete.
+        """
         self._notify = notify
 
     def _set_excinfo(self, excinfo):
@@ -300,7 +334,7 @@ class ResourceWorker(Worker):
     executes.
     """
 
-    def __init__(self, work_queue, resource):
+    def __init__(self, work_queue: queue.Queue, resource):
         super().__init__(work_queue)
         self._resource = resource
 
@@ -351,12 +385,7 @@ class ResourceThreadPoolExecutor(ThreadPoolExecutor):
         for worker in self._workers:
             worker.start()
 
-    def __repr__(self):
-        return "<ResourceThreadPoolExecutor ({} workers) at 0x{:012x}>".format(
-            len(self._workers), id(self)
-        )
-
-    def submit(self, future):
+    def submit(self, future: ResourceFuture) -> ResourceFuture:
         """
         Pass the ``task`` onto the workers'
         :ref:`uproot.source.futures.ResourceWorker.work_queue` as a
@@ -367,9 +396,7 @@ class ResourceThreadPoolExecutor(ThreadPoolExecutor):
         assert isinstance(future, ResourceFuture)
         if self.closed:
             raise OSError(
-                "resource is closed for file {}".format(
-                    self._workers[0].resource.file_path
-                )
+                f"resource is closed for file {self._workers[0].resource.file_path}"
             )
         self._work_queue.put(future)
         return future
@@ -382,7 +409,7 @@ class ResourceThreadPoolExecutor(ThreadPoolExecutor):
         self.__exit__(None, None, None)
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         """
         True if the :doc:`uproot.source.futures.ResourceWorker` threads have
         been stopped and their
@@ -399,3 +426,55 @@ class ResourceThreadPoolExecutor(ThreadPoolExecutor):
         for worker in self._workers:
             worker.resource.__exit__(exception_type, exception_value, traceback)
         self._closed = True
+
+
+##################### use-case 4: resources for I/O with trivial executor
+
+
+class ResourceTrivialExecutor(TrivialExecutor):
+    """
+    Args:
+        resource (:doc:`uproot.source.chunk.Resource`): Resource to
+            wrap as :doc:`uproot.source.futures.ResourceFuture` object.
+
+    A :doc:`uproot.source.futures.ResourceTrivialExecutor` is bound
+    to a resource, such as a file handle.
+    """
+
+    def __init__(self, resource):
+        self._resource = resource
+        self._closed = False
+
+    def submit(self, future: ResourceFuture) -> ResourceFuture:
+        """
+        Pass the ``task`` as a
+        :doc:`uproot.source.futures.ResourceFuture` so that it will be
+        executed with its :ref:`self._resource`.
+        """
+        assert isinstance(future, ResourceFuture)
+        if self.closed:
+            raise OSError(f"resource is closed for file {self._resource.file_path}")
+        future._run(self._resource)
+        return future
+
+    def close(self):
+        """
+        Stops the :doc:`uproot.source.futures.ResourceTrivialExecutor`
+        """
+        self.__exit__(None, None, None)
+
+    @property
+    def closed(self) -> bool:
+        """
+        True if the :doc:`uproot.source.futures.ResourceTrivialExecutor` has
+        been stopped.
+        """
+        return self._closed
+
+    def __enter__(self):
+        self._resource.__enter__()
+
+    def __exit__(self, exception_type, exception_value, traceback):
+        self.shutdown()
+        self._resource.__exit__(exception_type, exception_value, traceback)
+        self._closed = True
