@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines a physical layer for remote files, accessed via HTTP(S).
@@ -14,12 +14,14 @@ automatically falls back to :doc:`uproot.source.http.MultithreadedHTTPSource`.
 Despite the name, both sources support secure HTTPS (selected by URL scheme).
 """
 
+from __future__ import annotations
 
 import base64
+import http.client
 import queue
 import re
 import sys
-from http.client import HTTPConnection, HTTPSConnection
+import urllib.parse
 from urllib.parse import urlparse
 
 import uproot
@@ -27,7 +29,7 @@ import uproot.source.chunk
 import uproot.source.futures
 
 
-def make_connection(parsed_url, timeout):
+def make_connection(parsed_url: urllib.parse.ParseResult, timeout: float | None):
     """
     Args:
         parsed_url (``urllib.parse.ParseResult``): The URL to connect to, which
@@ -37,23 +39,21 @@ def make_connection(parsed_url, timeout):
     Creates a ``http.client.HTTPConnection`` or a ``http.client.HTTPSConnection``,
     depending on the URL scheme.
     """
+    from http.client import HTTPConnection, HTTPSConnection
+
     if parsed_url.scheme == "https":
-        return HTTPSConnection(
-            parsed_url.hostname, parsed_url.port, None, None, timeout
-        )
+        return HTTPSConnection(parsed_url.hostname, parsed_url.port, timeout=timeout)
 
     elif parsed_url.scheme == "http":
         return HTTPConnection(parsed_url.hostname, parsed_url.port, timeout)
 
     else:
         raise ValueError(
-            "unrecognized URL scheme for HTTP MultipartSource: {}".format(
-                parsed_url.scheme
-            )
+            f"unrecognized URL scheme for HTTP MultipartSource: {parsed_url.scheme}"
         )
 
 
-def full_path(parsed_url):
+def full_path(parsed_url) -> str:
     """
     Returns the ``parsed_url.path`` with ``"?"`` and the ``parsed_url.query``
     if it exists, just the path otherwise.
@@ -80,7 +80,7 @@ def basic_auth_headers(parsed_url):
     return ret
 
 
-def get_num_bytes(file_path, parsed_url, timeout):
+def get_num_bytes(file_path: str, parsed_url: urllib.parse.ParseResult, timeout) -> int:
     """
     Args:
         file_path (str): The URL to access as a raw string.
@@ -106,11 +106,9 @@ def get_num_bytes(file_path, parsed_url, timeout):
                 response = connection.getresponse()
                 break
         else:
-            raise OSError(
-                """remote server responded with status {} (redirect) without a 'location'
-for URL {}""".format(
-                    response.status, file_path
-                )
+            raise http.client.HTTPException(
+                f"""remote server responded with status {response.status} (redirect) without a 'location'
+for URL {file_path}"""
             )
 
     if response.status == 404:
@@ -119,11 +117,9 @@ for URL {}""".format(
 
     if response.status != 200:
         connection.close()
-        raise OSError(
-            """HTTP response was {}, rather than 200, in attempt to get file size
-in file {}""".format(
-                response.status, file_path
-            )
+        raise http.client.HTTPException(
+            f"""HTTP response was {response.status}, rather than 200, in attempt to get file size
+in file {file_path}"""
         )
 
     for k, x in response.getheaders():
@@ -132,11 +128,9 @@ in file {}""".format(
             return int(x)
     else:
         connection.close()
-        raise OSError(
-            """response headers did not include content-length: {}
-in file {}""".format(
-                dict(response.getheaders()), file_path
-            )
+        raise http.client.HTTPException(
+            f"""response headers did not include content-length: {dict(response.getheaders())}
+in file {file_path}"""
         )
 
 
@@ -160,7 +154,7 @@ class HTTPResource(uproot.source.chunk.Resource):
         self._auth_headers = basic_auth_headers(self._parsed_url)
 
     @property
-    def timeout(self):
+    def timeout(self) -> float | None:
         """
         The timeout in seconds or None.
         """
@@ -186,7 +180,7 @@ class HTTPResource(uproot.source.chunk.Resource):
     def __exit__(self, exception_type, exception_value, traceback):
         pass
 
-    def get(self, connection, start, stop):
+    def get(self, connection, start: int, stop: int) -> bytes:
         """
         Args:
             start (int): Seek position of the first byte to include.
@@ -216,20 +210,16 @@ class HTTPResource(uproot.source.chunk.Resource):
                     )
                     return self.get(redirect, start, stop)
 
-            raise OSError(
-                """remote server responded with status {} (redirect) without a 'location'
-for URL {}""".format(
-                    response.status, self._file_path
-                )
+            raise http.client.HTTPException(
+                f"""remote server responded with status {response.status} (redirect) without a 'location'
+for URL {self._file_path}"""
             )
 
         if response.status != 206:
             connection.close()
-            raise OSError(
-                """remote server responded with status {}, rather than 206 (range requests)
-for URL {}""".format(
-                    response.status, self._file_path
-                )
+            raise http.client.HTTPException(
+                f"""remote server responded with status {response.status}, rather than 206 (range requests)
+for URL {self._file_path}"""
             )
         try:
             return response.read()
@@ -237,7 +227,7 @@ for URL {}""".format(
             connection.close()
 
     @staticmethod
-    def future(source, start, stop):
+    def future(source: uproot.source.chunk.Source, start: int, stop: int):
         """
         Args:
             source (:doc:`uproot.source.http.HTTPSource` or :doc:`uproot.source.http.MultithreadedHTTPSource`): The
@@ -262,7 +252,13 @@ for URL {}""".format(
         return uproot.source.futures.ResourceFuture(task)
 
     @staticmethod
-    def multifuture(source, ranges, futures, results):
+    def multifuture(
+        source: uproot.source.chunk.Source,
+        range_header: dict,
+        ranges: list[(int, int)],
+        futures,
+        results,
+    ):
         """
         Args:
             source (:doc:`uproot.source.http.HTTPSource`): The data source.
@@ -285,49 +281,40 @@ for URL {}""".format(
         ``results`` and ``futures``. Subsequent attempts would immediately
         use the :ref:`uproot.source.http.HTTPSource.fallback`.
         """
-        connection = [make_connection(source.parsed_url, source.timeout)]
+        connection = make_connection(source.parsed_url, source.timeout)
 
-        range_strings = []
-        for start, stop in ranges:
-            range_strings.append(f"{start}-{stop - 1}")
-
-        connection[0].request(
+        connection.request(
             "GET",
             full_path(source.parsed_url),
-            headers=dict(
-                {"Range": "bytes=" + ", ".join(range_strings)}, **source.auth_headers
-            ),
+            headers=dict(**range_header, **source.auth_headers),
         )
 
         def task(resource):
+            nonlocal connection
             try:
-                response = connection[0].getresponse()
+                response = connection.getresponse()
 
                 if 300 <= response.status < 400:
-                    connection[0].close()
+                    connection.close()
 
                     for k, x in response.getheaders():
                         if k.lower() == "location":
                             redirect_url = urlparse(x)
-                            connection[0] = make_connection(
-                                redirect_url, source.timeout
-                            )
-                            connection[0].request(
+                            connection = make_connection(redirect_url, source.timeout)
+                            connection.request(
                                 "GET",
                                 full_path(redirect_url),
-                                headers=dict(
-                                    {"Range": "bytes=" + ", ".join(range_strings)},
+                                headers={
+                                    **range_header,
                                     **source.auth_headers,
-                                ),
+                                },
                             )
                             task(resource)
                             return
 
-                    raise OSError(
-                        """remote server responded with status {} (redirect) without a 'location'
-for URL {}""".format(
-                            response.status, source.file_path
-                        )
+                    raise http.client.HTTPException(
+                        f"""remote server responded with status {response.status} (redirect) without a 'location'
+for URL {source.file_path}"""
                     )
 
                 multipart_supported = resource.is_multipart_supported(ranges, response)
@@ -335,7 +322,9 @@ for URL {}""".format(
                 if not multipart_supported:
                     resource.handle_no_multipart(source, ranges, futures, results)
                 else:
-                    resource.handle_multipart(source, futures, results, response)
+                    resource.handle_multipart(
+                        source, futures, results, response, ranges
+                    )
 
             except Exception:
                 excinfo = sys.exc_info()
@@ -343,7 +332,7 @@ for URL {}""".format(
                     future._set_excinfo(excinfo)
 
             finally:
-                connection[0].close()
+                connection.close()
 
         return uproot.source.futures.ResourceFuture(task)
 
@@ -352,7 +341,9 @@ for URL {}""".format(
     )
     _content_range = re.compile(b"Content-Range: bytes ([0-9]+-[0-9]+)", re.I)
 
-    def is_multipart_supported(self, ranges, response):
+    def is_multipart_supported(
+        self, ranges: list[(int, int)], response: http.client.HTTPResponse
+    ) -> bool:
         """
         Helper function for :ref:`uproot.source.http.HTTPResource.multifuture`
         to check for multipart GET support.
@@ -369,7 +360,13 @@ for URL {}""".format(
         else:
             return True
 
-    def handle_no_multipart(self, source, ranges, futures, results):
+    def handle_no_multipart(
+        self,
+        source: uproot.source.chunk.Source,
+        ranges: list[(int, int)],
+        futures,
+        results,
+    ):
         """
         Helper function for :ref:`uproot.source.http.HTTPResource.multifuture`
         to handle a lack of multipart GET support.
@@ -384,7 +381,14 @@ for URL {}""".format(
             results[chunk.start, chunk.stop] = chunk.raw_data
             futures[chunk.start, chunk.stop]._run(self)
 
-    def handle_multipart(self, source, futures, results, response):
+    def handle_multipart(
+        self,
+        source: uproot.source.chunk.Source,
+        futures,
+        results,
+        response: http.client.HTTPResponse,
+        ranges: list[(int, int)],
+    ):
         """
         Helper function for :ref:`uproot.source.http.HTTPResource.multifuture`
         to handle the multipart GET response.
@@ -394,17 +398,15 @@ for URL {}""".format(
         else:
             response_buffer = _ResponseBuffer(response)
 
+        original_futures = dict(futures)
+
         num_found = 0
         while len(futures) > 0:
             range_string, size = self.next_header(response_buffer)
-            if range_string is None:
-                raise OSError(
-                    """found {} of {} expected headers in HTTP multipart
-for URL {}""".format(
-                        num_found, len(futures), self._file_path
-                    )
-                )
             num_found += 1
+            if range_string is None:
+                self.handle_no_multipart(source, ranges, original_futures, results)
+                return
 
             start, last = range_string.split(b"-")
             start, last = int(start), int(last)
@@ -414,12 +416,10 @@ for URL {}""".format(
             data = response_buffer.read(length)
 
             if len(data) != length:
-                raise OSError(
-                    """wrong chunk length {} (expected {}) for byte range {} "
+                raise http.client.HTTPException(
+                    f"""wrong chunk length {len(data)} (expected {length}) for byte range {range_string.decode()!r} "
                     "in HTTP multipart
-for URL {}""".format(
-                        len(data), length, repr(range_string.decode()), self._file_path
-                    )
+for URL {self._file_path}"""
                 )
 
             found = futures.pop((start, stop), None)
@@ -440,14 +440,12 @@ for URL {}""".format(
                     else:
                         range_string = range_string.decode("utf-8", "surrogateescape")
                         expecting = ", ".join(f"{a}-{b - 1}" for a, b in futures)
-                        raise OSError(
-                            """unrecognized byte range in headers of HTTP multipart: {}
+                        raise http.client.HTTPException(
+                            f"""unrecognized byte range in headers of HTTP multipart: {range_string!r}
 
-    expecting: {}
+    expecting: {expecting}
 
-for URL {}""".format(
-                                repr(range_string), expecting, self._file_path
-                            )
+for URL {self._file_path}"""
                         )
                     subdata = data[
                         now - start : now + future_stop - future_start - start
@@ -480,7 +478,7 @@ for URL {}""".format(
         return range_string, size
 
     @staticmethod
-    def partfuture(results, start, stop):
+    def partfuture(results, start: int, stop: int):
         """
         Returns a :doc:`uproot.source.futures.ResourceFuture` to simply select
         the ``(start, stop)`` item from the ``results`` dict.
@@ -503,7 +501,7 @@ class _ResponseBuffer:
         self.already_read = b""
         self.stream = stream
 
-    def read(self, length):
+    def read(self, length: int):
         if length < len(self.already_read):
             out = self.already_read[:length]
             self.already_read = self.already_read[length:]
@@ -517,7 +515,7 @@ class _ResponseBuffer:
         else:
             return self.stream.read(length)
 
-    def readline(self):
+    def readline(self) -> bytes:
         while True:
             try:
                 index = self.already_read.index(b"\n")
@@ -533,7 +531,7 @@ class HTTPSource(uproot.source.chunk.Source):
     """
     Args:
         file_path (str): A URL of the file to open.
-        options: Must include ``"num_fallback_workers"`` and ``"timeout"``.
+        options: Must include ``"num_fallback_workers"``, ``"use_threads"``, and ``"timeout"``.
 
     A :doc:`uproot.source.chunk.Source` that first attempts an HTTP(S)
     multipart GET, but if the server doesn't support it, it falls back to many
@@ -548,25 +546,40 @@ class HTTPSource(uproot.source.chunk.Source):
 
     ResourceClass = HTTPResource
 
-    def __init__(self, file_path, **options):
+    def __init__(self, file_path: str, **options):
+        options = dict(uproot.reading.open.defaults, **options)
+
         self._num_fallback_workers = options["num_fallback_workers"]
         self._timeout = options["timeout"]
         self._num_requests = 0
         self._num_requested_chunks = 0
         self._num_requested_bytes = 0
 
+        self._use_threads = options["use_threads"]
         self._file_path = file_path
         self._num_bytes = None
 
         self._fallback = None
-        self._fallback_options = dict(options)
+        self._fallback_options = options.copy()
         self._fallback_options["num_workers"] = self._num_fallback_workers
+
+        self._http_max_header_bytes = options["http_max_header_bytes"]
+
+        # Parse the URL here, so that we can expose these fields
+        self._parsed_url = urlparse(file_path)
+        self._auth_headers = basic_auth_headers(self._parsed_url)
+
         self._open()
 
     def _open(self):
-        self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
-            [HTTPResource(self._file_path, self._timeout)]
-        )
+        if self._use_threads:
+            self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
+                [HTTPResource(self._file_path, self._timeout)]
+            )
+        else:
+            self._executor = uproot.source.futures.ResourceTrivialExecutor(
+                HTTPResource(self._file_path, self._timeout)
+            )
 
     def __getstate__(self):
         state = dict(self.__dict__)
@@ -582,13 +595,11 @@ class HTTPSource(uproot.source.chunk.Source):
         if len(self._file_path) > 10:
             path = repr("..." + self._file_path[-10:])
         fallback = ""
-        if self._fallback is not None:
+        if getattr(self, "_fallback", None) is not None:
             fallback = " with fallback"
-        return "<{} {}{} at 0x{:012x}>".format(
-            type(self).__name__, path, fallback, id(self)
-        )
+        return f"<{type(self).__name__} {path}{fallback} at 0x{id(self):012x}>"
 
-    def chunk(self, start, stop):
+    def chunk(self, start: int, stop: int) -> uproot.source.chunk.Chunk:
         self._num_requests += 1
         self._num_requested_chunks += 1
         self._num_requested_bytes += stop - start
@@ -598,28 +609,64 @@ class HTTPSource(uproot.source.chunk.Source):
         self._executor.submit(future)
         return chunk
 
-    def chunks(self, ranges, notifications):
+    def chunks(
+        self,
+        ranges: list[(int, int)],
+        notifications: queue.Queue,
+    ) -> list[uproot.source.chunk.Chunk]:
         if self._fallback is None:
             self._num_requests += 1
             self._num_requested_chunks += len(ranges)
             self._num_requested_bytes += sum(stop - start for start, stop in ranges)
-
-            futures = {}
-            results = {}
             chunks = []
-            for start, stop in ranges:
-                partfuture = self.ResourceClass.partfuture(results, start, stop)
-                futures[start, stop] = partfuture
-                results[start, stop] = None
-                chunk = uproot.source.chunk.Chunk(self, start, stop, partfuture)
-                partfuture._set_notify(
-                    uproot.source.chunk.notifier(chunk, notifications)
-                )
-                chunks.append(chunk)
 
-            self._executor.submit(
-                self.ResourceClass.multifuture(self, ranges, futures, results)
-            )
+            def set_futures_and_results(ranges):
+                futures = {}
+                results = {}
+
+                for start, stop in ranges:
+                    partfuture = self.ResourceClass.partfuture(results, start, stop)
+                    futures[start, stop] = partfuture
+                    results[start, stop] = None
+                    chunk = uproot.source.chunk.Chunk(self, start, stop, partfuture)
+                    partfuture._set_notify(
+                        uproot.source.chunk.notifier(chunk, notifications)
+                    )
+                    chunks.append(chunk)
+
+                return futures, results
+
+            i, j = 1, 0
+            range_header = {"Range": "bytes=" + f"{ranges[0][0]}-{ranges[0][1] - 1}"}
+            last_batch_appended = False
+
+            while i < len(ranges):
+                new_range_to_append = ", " + f"{ranges[i][0]}-{ranges[i][1] - 1}"
+                if len(range_header["Range"]) < self._http_max_header_bytes - len(
+                    new_range_to_append
+                ):
+                    range_header["Range"] += new_range_to_append
+                    last_batch_appended = False
+                else:
+                    futures, results = set_futures_and_results(ranges[j : j + i])
+                    self._executor.submit(
+                        self.ResourceClass.multifuture(
+                            self, range_header, ranges[j : j + i], futures, results
+                        )
+                    )
+                    j += i
+                    range_header = {"Range": "bytes=" + new_range_to_append[1:]}
+                    last_batch_appended = True
+                i += 1
+
+            if i == len(ranges) and not last_batch_appended:
+                futures, results = set_futures_and_results(ranges[j:])
+                self._executor.submit(
+                    self.ResourceClass.multifuture(
+                        self, range_header, ranges[j : j + i], futures, results
+                    )
+                )
+
             return chunks
 
         else:
@@ -634,7 +681,7 @@ class HTTPSource(uproot.source.chunk.Source):
         return self._executor
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         return self._executor.closed
 
     def __enter__(self):
@@ -653,7 +700,7 @@ class HTTPSource(uproot.source.chunk.Source):
         return self._timeout
 
     @property
-    def num_bytes(self):
+    def num_bytes(self) -> int:
         if self._num_bytes is None:
             self._num_bytes = get_num_bytes(
                 self._file_path, self.parsed_url, self._timeout
@@ -665,14 +712,14 @@ class HTTPSource(uproot.source.chunk.Source):
         """
         A ``urllib.parse.ParseResult`` version of the ``file_path``.
         """
-        return self._executor.workers[0].resource.parsed_url
+        return self._parsed_url
 
     @property
     def auth_headers(self):
         """
         Dict containing auth headers, if any
         """
-        return self._executor.workers[0].resource.auth_headers
+        return self._auth_headers
 
     @property
     def fallback(self):
@@ -687,8 +734,7 @@ class HTTPSource(uproot.source.chunk.Source):
 
     def _set_fallback(self):
         self._fallback = MultithreadedHTTPSource(
-            self._file_path,
-            **self._fallback_options,  # NOTE: a comma after **fallback_options breaks Python 2
+            self._file_path, **self._fallback_options
         )
 
 
@@ -696,7 +742,7 @@ class MultithreadedHTTPSource(uproot.source.chunk.MultithreadedSource):
     """
     Args:
         file_path (str): A URL of the file to open.
-        options: Must include ``"num_workers"`` and ``"timeout"``.
+        options: Must include ``"num_workers"``, ``"use_threads"``, and ``"timeout"``.
 
     A :doc:`uproot.source.chunk.MultithreadedSource` that manages many
     :doc:`uproot.source.http.HTTPResource` objects.
@@ -705,19 +751,43 @@ class MultithreadedHTTPSource(uproot.source.chunk.MultithreadedSource):
     ResourceClass = HTTPResource
 
     def __init__(self, file_path, **options):
-        num_workers = options["num_workers"]
-        timeout = options["timeout"]
         self._num_requests = 0
         self._num_requested_chunks = 0
         self._num_requested_bytes = 0
+        self._use_threads = options["use_threads"]
+        self._num_workers = options["num_workers"]
 
         self._file_path = file_path
         self._num_bytes = None
-        self._timeout = timeout
+        self._timeout = options["timeout"]
 
-        self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
-            [HTTPResource(file_path, timeout) for x in range(num_workers)]
-        )
+        # Parse the URL here, so that we can expose these fields
+        self._parsed_url = urlparse(file_path)
+        self._auth_headers = basic_auth_headers(self._parsed_url)
+
+        self._open()
+
+    def __getstate__(self):
+        state = dict(self.__dict__)
+        state.pop("_executor")
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__ = state
+        self._open()
+
+    def _open(self):
+        if self._use_threads:
+            self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
+                [
+                    HTTPResource(self._file_path, self._timeout)
+                    for _ in range(self._num_workers)
+                ]
+            )
+        else:
+            self._executor = uproot.source.futures.ResourceTrivialExecutor(
+                HTTPResource(self._file_path, self._timeout)
+            )
 
     @property
     def timeout(self):
@@ -727,7 +797,7 @@ class MultithreadedHTTPSource(uproot.source.chunk.MultithreadedSource):
         return self._timeout
 
     @property
-    def num_bytes(self):
+    def num_bytes(self) -> int:
         if self._num_bytes is None:
             self._num_bytes = get_num_bytes(
                 self._file_path, self.parsed_url, self._timeout
@@ -739,11 +809,11 @@ class MultithreadedHTTPSource(uproot.source.chunk.MultithreadedSource):
         """
         A ``urllib.parse.ParseResult`` version of the ``file_path``.
         """
-        return self._executor.workers[0].resource.parsed_url
+        return self._parsed_url
 
     @property
     def auth_headers(self):
         """
         Dict containing auth headers, if any
         """
-        return self._executor.workers[0].resource.auth_headers
+        return self._auth_headers
