@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines a physical layer for remote files, accessed via the XRootD protocol.
@@ -10,7 +10,10 @@ support vector-read requests; if not, it automatically falls back to
 :doc:`uproot.source.xrootd.MultithreadedXRootDSource`.
 """
 
+from __future__ import annotations
 
+import contextlib
+import queue
 import sys
 
 import uproot
@@ -60,7 +63,10 @@ def get_server_config(file):
         XRootD_client.flags.QueryCode.CONFIG, "readv_iov_max readv_ior_max"
     )
     if status.error:
-        raise OSError(status.message)
+        if status.code in (206,):
+            raise TimeoutError(status.message)
+        else:
+            raise OSError(status.message)
 
     # Result is something like b'178956968\n2097136\n'
     readv_iov_max, readv_ior_max = map(int, result.split(b"\n", 1))
@@ -68,6 +74,11 @@ def get_server_config(file):
     return readv_iov_max, readv_ior_max
 
 
+@contextlib.contextmanager
+def trivial_resource():
+    yield
+
+
 class XRootDResource(uproot.source.chunk.Resource):
     """
     Args:
@@ -77,7 +88,7 @@ class XRootDResource(uproot.source.chunk.Resource):
     A :doc:`uproot.source.chunk.Resource` for XRootD connections.
     """
 
-    def __init__(self, file_path, timeout):
+    def __init__(self, file_path: str, timeout: float | None):
         self._file_path = file_path
         self._timeout = timeout
         self._open()
@@ -100,7 +111,7 @@ class XRootDResource(uproot.source.chunk.Resource):
         self.__dict__ = state
         self._open()
 
-    def _xrd_timeout(self):
+    def _xrd_timeout(self) -> int:
         if self._timeout is None:
             return 0
         else:
@@ -111,19 +122,17 @@ class XRootDResource(uproot.source.chunk.Resource):
 
         # https://github.com/xrootd/xrootd/blob/8e91462e76ab969720b40fc324714b84e0b4bd42/src/XrdCl/XrdClStatus.hh#L47-L103
         # https://github.com/xrootd/xrootd/blob/250eced4d3787c2ac5be2c8c922134153bbf7f08/src/XrdCl/XrdClStatus.cc#L34-L74
-        if status.code == 101 or status.code == 304 or status.code == 400:
+        if status.code in (101, 304, 400):
             raise uproot._util._file_not_found(self._file_path, status.message)
-
-        else:
-            raise OSError(
-                """XRootD error: {}
-in file {}""".format(
-                    status.message, self._file_path
-                )
+        elif status.code in (206,):
+            raise TimeoutError(
+                f"XRootD error: {status.message} in file {self._file_path}"
             )
+        else:
+            raise OSError(f"XRootD error: {status.message} in file {self._file_path}")
 
     @property
-    def timeout(self):
+    def timeout(self) -> float | None:
         """
         The timeout in seconds or None.
         """
@@ -137,14 +146,14 @@ in file {}""".format(
         return self._file
 
     @property
-    def num_bytes(self):
+    def num_bytes(self) -> int:
         status, info = self._file.stat(self._xrd_timeout())
         if status.error:
             self._xrd_error(status)
         return info.size
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         return not self._file.is_open()
 
     def __enter__(self):
@@ -153,7 +162,7 @@ in file {}""".format(
     def __exit__(self, exception_type, exception_value, traceback):
         self._file.close(timeout=self._xrd_timeout())
 
-    def get(self, start, stop):
+    def get(self, start: int, stop: int) -> bytes:
         """
         Args:
             start (int): Seek position of the first byte to include.
@@ -170,7 +179,7 @@ in file {}""".format(
         return data
 
     @staticmethod
-    def future(source, start, stop):
+    def future(source: uproot.source.chunk.Source, start: int, stop: int):
         """
         Args:
             source (:doc:`uproot.source.xrootd.MultithreadedXRootDSource`): The
@@ -190,7 +199,7 @@ in file {}""".format(
         return uproot.source.futures.ResourceFuture(task)
 
     @staticmethod
-    def partfuture(results, start, stop):
+    def partfuture(results, start: int, stop: int):
         """
         Returns a :doc:`uproot.source.futures.ResourceFuture` to simply select
         the ``(start, stop)`` item from the ``results`` dict.
@@ -251,7 +260,7 @@ class XRootDSource(uproot.source.chunk.Source):
     """
     Args:
         file_path (str): A URL of the file to open.
-        options: Must include ``"timeout"``, ``"max_num_elements"`` and ``"num_workers"``
+        options: Must include ``"timeout"``, ``"max_num_elements"``, ``"use_threads"``, and ``"num_workers"``
 
     A :doc:`uproot.source.chunk.Source` that uses XRootD's vector-read
     to get many chunks in one request.
@@ -259,9 +268,12 @@ class XRootDSource(uproot.source.chunk.Source):
 
     ResourceClass = XRootDResource
 
-    def __init__(self, file_path, **options):
+    def __init__(self, file_path: str, **options):
+        options = dict(uproot.reading.open.defaults, **options)
+
         self._timeout = options["timeout"]
         self._desired_max_num_elements = options["max_num_elements"]
+        self._use_threads = options["use_threads"]
         self._num_workers = options["num_workers"]
         self._num_requests = 0
         self._num_requested_chunks = 0
@@ -276,9 +288,14 @@ class XRootDSource(uproot.source.chunk.Source):
 
         # this ThreadPool does not need a resource, it's only used to submit
         # futures that wait for chunks that have been split to merge them.
-        self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
-            [None for i in range(self._num_workers)]
-        )
+        if self._use_threads:
+            self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
+                [trivial_resource() for _ in range(self._num_workers)]
+            )
+        else:
+            self._executor = uproot.source.futures.ResourceTrivialExecutor(
+                trivial_resource()
+            )
 
         self._max_num_elements, self._max_element_size = get_server_config(
             self._resource.file
@@ -303,7 +320,7 @@ class XRootDSource(uproot.source.chunk.Source):
             path = repr("..." + self._file_path[-10:])
         return f"<{type(self).__name__} {path} at 0x{id(self):012x}>"
 
-    def chunk(self, start, stop):
+    def chunk(self, start: int, stop: int) -> uproot.source.chunk.Chunk:
         self._num_requests += 1
         self._num_requested_chunks += 1
         self._num_requested_bytes += stop - start
@@ -312,7 +329,9 @@ class XRootDSource(uproot.source.chunk.Source):
         future = uproot.source.futures.TrivialFuture(data)
         return uproot.source.chunk.Chunk(self, start, stop, future)
 
-    def chunks(self, ranges, notifications):
+    def chunks(
+        self, ranges: list[(int, int)], notifications: queue.Queue
+    ) -> list[uproot.source.chunk.Chunk]:
         self._num_requests += 1
         self._num_requested_chunks += len(ranges)
         self._num_requested_bytes += sum(stop - start for start, stop in ranges)
@@ -325,7 +344,9 @@ class XRootDSource(uproot.source.chunk.Source):
         # this is to track which requests were split into smaller ranges and have to be merged
         sub_ranges = {}
 
-        def add_request_range(start, length, sub_ranges_list):
+        def add_request_range(
+            start: int, length: int, sub_ranges_list: list[(int, int)]
+        ):
             if len(all_request_ranges[-1]) >= self._max_num_elements:
                 all_request_ranges.append([])
             all_request_ranges[-1].append((start, length))
@@ -398,7 +419,7 @@ class XRootDSource(uproot.source.chunk.Source):
         return self._resource
 
     @property
-    def timeout(self):
+    def timeout(self) -> float | None:
         """
         The timeout in seconds or None.
         """
@@ -412,7 +433,7 @@ class XRootDSource(uproot.source.chunk.Source):
         return self._resource.file
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         return self._resource.closed
 
     def __enter__(self):
@@ -423,7 +444,7 @@ class XRootDSource(uproot.source.chunk.Source):
         self._resource.__exit__(exception_type, exception_value, traceback)
 
     @property
-    def num_bytes(self):
+    def num_bytes(self) -> int:
         if self._num_bytes is None:
             self._num_bytes = self._resource.num_bytes
         return self._num_bytes
@@ -433,7 +454,7 @@ class MultithreadedXRootDSource(uproot.source.chunk.MultithreadedSource):
     """
     Args:
         file_path (str): A URL of the file to open.
-        options: Must include ``"num_workers"`` and ``"timeout"``.
+        options: Must include ``"num_workers"``, ``"use_threads"``, and ``"timeout"``.
 
     A :doc:`uproot.source.chunk.MultithreadedSource` that manages many
     :doc:`uproot.source.xrootd.XRootDResource` objects.
@@ -441,9 +462,12 @@ class MultithreadedXRootDSource(uproot.source.chunk.MultithreadedSource):
 
     ResourceClass = XRootDResource
 
-    def __init__(self, file_path, **options):
+    def __init__(self, file_path: str, **options):
+        options = dict(uproot.reading.open.defaults, **options)
+
         self._num_workers = options["num_workers"]
         self._timeout = options["timeout"]
+        self._use_threads = options["use_threads"]
         self._num_requests = 0
         self._num_requested_chunks = 0
         self._num_requested_bytes = 0
@@ -453,12 +477,17 @@ class MultithreadedXRootDSource(uproot.source.chunk.MultithreadedSource):
         self._open()
 
     def _open(self):
-        self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
-            [
+        if self._use_threads:
+            self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
+                [
+                    XRootDResource(self._file_path, self._timeout)
+                    for _ in range(self._num_workers)
+                ]
+            )
+        else:
+            self._executor = uproot.source.futures.ResourceTrivialExecutor(
                 XRootDResource(self._file_path, self._timeout)
-                for x in range(self._num_workers)
-            ]
-        )
+            )
 
     def __getstate__(self):
         state = dict(self.__dict__)
@@ -470,14 +499,18 @@ class MultithreadedXRootDSource(uproot.source.chunk.MultithreadedSource):
         self._open()
 
     @property
-    def timeout(self):
+    def timeout(self) -> float | None:
         """
         The timeout in seconds or None.
         """
         return self._timeout
 
     @property
-    def num_bytes(self):
+    def num_bytes(self) -> int:
         if self._num_bytes is None:
-            self._num_bytes = self._executor.workers[0].resource.num_bytes
+            if hasattr(self._executor, "workers"):
+                self._num_bytes = self._executor.workers[0].resource.num_bytes
+            else:
+                self._num_bytes = self._executor._resource.num_bytes
+
         return self._num_bytes
