@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This is an internal module defining the "cascading" file writer, which is like a
@@ -18,7 +18,7 @@ as well, which might be in another TDirectory if it's a subdirectory, etc. If th
 updates were not organized, redundant overwrites or even infinite cycles could occur.
 Thus, the structure of the cascade tree must be carefully laid out.
 """
-
+from __future__ import annotations
 
 import datetime
 import math
@@ -36,7 +36,6 @@ import uproot.serialization
 import uproot.source.chunk
 import uproot.source.cursor
 import uproot.streamers
-import uproot.writing._cascadetree
 
 
 class CascadeLeaf:
@@ -152,11 +151,7 @@ class String(CascadeLeaf):
         self._serialization = uproot.serialization.string(self._string)
 
     def __repr__(self):
-        return "{}({}, {})".format(
-            type(self).__name__,
-            self._location,
-            repr(self._string),
-        )
+        return f"{type(self).__name__}({self._location}, {self._string!r})"
 
     @property
     def string(self):
@@ -223,18 +218,7 @@ class Key(CascadeLeaf):
         self._big = big
 
     def __repr__(self):
-        return "{}({}, {}, {}, {}, {}, {}, {}, {}, {})".format(
-            type(self).__name__,
-            self._location,
-            self._uncompressed_bytes,
-            self._compressed_bytes,
-            repr(self._classname),
-            repr(self._name),
-            repr(self._title),
-            self._cycle,
-            self._parent_location,
-            self._seek_location,
-        )
+        return f"{type(self).__name__}({self._location}, {self._uncompressed_bytes}, {self._compressed_bytes}, {self._classname!r}, {self._name!r}, {self._title!r}, {self._cycle}, {self._parent_location}, {self._seek_location})"
 
     @property
     def allocation(self):
@@ -430,19 +414,13 @@ class Key(CascadeLeaf):
 
         if version != cls.class_version:
             raise ValueError(
-                "Uproot can't read TKey version {} for writing, only version {}{}".format(
-                    version,
-                    cls.class_version,
-                    in_path,
-                )
+                f"Uproot can't read TKey version {version} for writing, only version {cls.class_version}{in_path}"
             )
 
         assert 0 < fNbytes <= fKeylen + fObjlen
         assert fCycle > 0
         if not is_directory_key:
-            assert fSeekKey == location, "fSeekKey {} location {}".format(
-                fSeekKey, location
-            )
+            assert fSeekKey == location, f"fSeekKey {fSeekKey} location {location}"
             fSeekKey = None
 
         classname, position = String.deserialize(
@@ -485,12 +463,7 @@ class FreeSegmentsData(CascadeLeaf):
         self._end = end
 
     def __repr__(self):
-        return "{}({}, {}, {})".format(
-            type(self).__name__,
-            self._location,
-            self._slices,
-            self._end,
-        )
+        return f"{type(self).__name__}({self._location}, {self._slices}, {self._end})"
 
     @property
     def slices(self):
@@ -593,11 +566,7 @@ class FreeSegmentsData(CascadeLeaf):
 
             if version != cls.class_version:
                 raise ValueError(
-                    "Uproot can't read TFree version {} for writing, only version {}{}".format(
-                        version,
-                        cls.class_version,
-                        in_path,
-                    )
+                    f"Uproot can't read TFree version {version} for writing, only version {cls.class_version}{in_path}"
                 )
 
             slices.append((fFirst, fLast + 1))
@@ -641,12 +610,7 @@ class FreeSegments(CascadeNode):
         self._fileheader = fileheader
 
     def __repr__(self):
-        return "{}({}, {}, {})".format(
-            type(self).__name__,
-            self._key,
-            self._data,
-            self._fileheader,
-        )
+        return f"{type(self).__name__}({self._key}, {self._data}, {self._fileheader})"
 
     @property
     def key(self):
@@ -712,9 +676,7 @@ class FreeSegments(CascadeNode):
             if start <= original_start < stop or start < original_stop <= stop:
                 raise RuntimeError(
                     "segment of data to release overlaps one already marked as free: "
-                    "releasing [{}, {}) but [{}, {}) is free".format(
-                        original_start, original_stop, start, stop
-                    )
+                    f"releasing [{original_start}, {original_stop}) but [{start}, {stop}) is free"
                 )
 
         for i in range(len(slices) - 1):
@@ -740,7 +702,7 @@ class FreeSegments(CascadeNode):
                 )
 
         # The FreeSegments record will have to grow.
-        return tuple(sorted(slices + ((original_start, original_stop),)))
+        return tuple(sorted((*slices, (original_start, original_stop))))
 
     @staticmethod
     def _slices_bytes(slices):
@@ -817,9 +779,7 @@ class TListHeader(CascadeLeaf):
         self._num_entries = num_entries
 
     def __repr__(self):
-        return "{}({}, {}, {})".format(
-            type(self).__name__, self._location, self._data_bytes, self._num_entries
-        )
+        return f"{type(self).__name__}({self._location}, {self._data_bytes}, {self._num_entries})"
 
     @property
     def data_bytes(self):
@@ -879,13 +839,7 @@ class RawStreamerInfo(CascadeLeaf):
         return self._class_version
 
     def __repr__(self):
-        return "{}({}, {}, {}, {})".format(
-            type(self).__name__,
-            self._location,
-            self._serialization,
-            repr(self._name),
-            self._class_version,
-        )
+        return f"{type(self).__name__}({self._location}, {self._serialization}, {self._name!r}, {self._class_version})"
 
     def copy_to(self, location):
         return RawStreamerInfo(
@@ -910,11 +864,7 @@ class RawTListOfStrings(CascadeLeaf):
         self._serialization = serialization
 
     def __repr__(self):
-        return "{}({}, {})".format(
-            type(self).__name__,
-            self._location,
-            self._serialization,
-        )
+        return f"{type(self).__name__}({self._location}, {self._serialization})"
 
     def copy_to(self, location):
         return RawTListOfStrings(location, self._serialization)
@@ -944,14 +894,7 @@ class TListOfStreamers(CascadeNode):
         }
 
     def __repr__(self):
-        return "{}({}, {}, {}, {}, {})".format(
-            type(self).__name__,
-            self._allocation,
-            self._key,
-            self._header,
-            self._rawstreamers,
-            self._freesegments,
-        )
+        return f"{type(self).__name__}({self._allocation}, {self._key}, {self._header}, {self._rawstreamers}, {self._freesegments})"
 
     @property
     def allocation(self):
@@ -1176,11 +1119,8 @@ class DirectoryData(CascadeLeaf):
         self._keys = keys
 
     def __repr__(self):
-        return "{}({}, {}, {})".format(
-            type(self).__name__,
-            self._location,
-            self._allocation,
-            self._keys,
+        return (
+            f"{type(self).__name__}({self._location}, {self._allocation}, {self._keys})"
         )
 
     @property
@@ -1388,16 +1328,7 @@ class DirectoryHeader(CascadeLeaf):
         self._modified_on = self._created_on
 
     def __repr__(self):
-        return "{}({}, {}, {}, {}, {}, {}, {})".format(
-            type(self).__name__,
-            self._location,
-            self._begin_location,
-            self._begin_num_bytes,
-            self._data_location,
-            self._data_num_bytes,
-            self._parent_location,
-            repr(self._uuid),
-        )
+        return f"{type(self).__name__}({self._location}, {self._begin_location}, {self._begin_num_bytes}, {self._data_location}, {self._data_num_bytes}, {self._parent_location}, {self._uuid!r})"
 
     @property
     def begin_location(self):
@@ -1531,11 +1462,7 @@ class DirectoryHeader(CascadeLeaf):
 
         if version != cls.class_version:
             raise ValueError(
-                "Uproot can't read TDirectory version {} for writing, only version {}{}".format(
-                    version,
-                    cls.class_version,
-                    in_path,
-                )
+                f"Uproot can't read TDirectory version {version} for writing, only version {cls.class_version}{in_path}"
             )
 
         # TUUID version 1
@@ -1596,10 +1523,7 @@ class Directory(CascadeNode):
         replaces=None,
         big=None,
     ):
-        if replaces is None:
-            cycle = self._data.next_cycle(name)
-        else:
-            cycle = replaces.cycle
+        cycle = self._data.next_cycle(name) if replaces is None else replaces.cycle
 
         classname_asbytes = classname.encode(errors="surrogateescape")
         name_asbytes = name.encode(errors="surrogateescape")
@@ -1662,6 +1586,7 @@ class Directory(CascadeNode):
                 )
                 self._reallocate_data(requested_num_bytes)
                 next_key = key.copy_to(self._data.next_location)
+            next_key._location = self._data.next_location
             self._data.add_key(next_key, incremental=True)
 
         else:
@@ -1763,8 +1688,8 @@ class Directory(CascadeNode):
                 int(math.ceil(1.5 * (self._data.allocation + next_key.num_bytes + 8)))
             )
             next_key = subdirectory_key.copy_to(self._data.next_location)
+        next_key._location = self._data.next_location
         self._data.add_key(next_key, incremental=True)
-
         self._header.modified_on = datetime.datetime.now()
 
         if flush:
@@ -1786,6 +1711,8 @@ class Directory(CascadeNode):
         initial_basket_capacity,
         resize_factor,
     ):
+        import uproot.writing._cascadetree
+
         tree = uproot.writing._cascadetree.Tree(
             self,
             name,
@@ -1800,6 +1727,65 @@ class Directory(CascadeNode):
         tree.write_anew(sink)
         return tree
 
+    def add_rntuple(self, sink, name, title, akform):
+        import uproot.writing._cascadentuple
+
+        anchor = uproot.writing._cascadentuple.NTuple_Anchor(
+            None, 0, 0, 48, None, None, None, None, None, None, 0
+        )
+
+        header = uproot.writing._cascadentuple.NTuple_Header(None, name, "", akform)
+
+        footer = uproot.writing._cascadentuple.NTuple_Footer(
+            None, 0, header._crc32, akform
+        )
+
+        # the empty page list is hard-coded bytes which represents:
+        # 0                   1                   2                   3
+        # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+        # |        Envelope Version       |        Minimum Version        |
+        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+        # |                             Size                            |T|
+        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+        # |           Number of Items (for list frames)           |Reserv.|
+        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+        # |                         FRAME PAYLOAD                         |
+        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+        # |                             CRC32                             |
+        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+        #
+        # - Envelope Version = 1 (0x0100)
+        # - Minimum Version = 1  (0x0100)
+        # - Size = -8 (0xf8ffffff) [value is negative because this is a list]
+        # - Number of Items = 0 (0x00000000) [empty list]
+        # - FRAME PAYLOAD = empty [because number of items is 0]
+        # - CRC32 = 2678769841
+        # manually calculate CRC32:
+
+        # In [1]: zlib.crc32(b'\x01\x00\x01\x00\xf8\xff\xff\xff\00\00\00\00')
+        # Out[1]: 2678769841
+        # In [2]: np.array([177, 200, 170, 159], dtype=np.uint8).view("uint32")
+        # Out[2]: array([2678769841], dtype=uint32)
+
+        empty_page_list_bytes = numpy.array(
+            [1, 0, 1, 0, 248, 255, 255, 255, 0, 0, 0, 0, 177, 200, 170, 159],
+            dtype=numpy.uint8,
+        )
+        offset = self._freesegments.allocate(16)
+        footer.cluster_group_record_frames[0].page_list_envlink.locator = (
+            uproot.writing._cascadentuple.NTuple_Locator(16, offset)
+        )
+
+        ntuple = uproot.writing._cascadentuple.NTuple(
+            self, name, title, akform, self._freesegments, header, footer, [], anchor
+        )
+
+        sink.write(offset, empty_page_list_bytes)
+        ntuple.write(sink)
+        sink.flush()
+        return ntuple
+
 
 class RootDirectory(Directory):
     """
@@ -1820,16 +1806,7 @@ class RootDirectory(Directory):
         self._freesegments = freesegments
 
     def __repr__(self):
-        return "{}({}, {}, {}, {}, {}, {}, {})".format(
-            type(self).__name__,
-            self._key,
-            self._name,
-            self._title,
-            self._header,
-            self._datakey,
-            self._data,
-            self._freesegments,
-        )
+        return f"{type(self).__name__}({self._key}, {self._name}, {self._title}, {self._header}, {self._datakey}, {self._data}, {self._freesegments})"
 
     @property
     def key(self):
@@ -1904,15 +1881,7 @@ class SubDirectory(Directory):
         self._freesegments = freesegments
 
     def __repr__(self):
-        return "{}({}, {}, {}, {}, {}, {})".format(
-            type(self).__name__,
-            self._key,
-            self._header,
-            self._datakey,
-            self._data,
-            self._parent,
-            self._freesegments,
-        )
+        return f"{type(self).__name__}({self._key}, {self._header}, {self._datakey}, {self._data}, {self._parent}, {self._freesegments})"
 
     @property
     def key(self):
@@ -2001,18 +1970,7 @@ class FileHeader(CascadeLeaf):
         self._begin = 100
 
     def __repr__(self):
-        return "{}({}, {}, {}, {}, {}, {}, {}, {}, {})".format(
-            type(self).__name__,
-            self._end,
-            self._free_location,
-            self._free_num_bytes,
-            self._free_num_slices,
-            self._begin_num_bytes,
-            repr(self._compression),
-            self._info_location,
-            self._info_num_bytes,
-            repr(self._uuid),
-        )
+        return f"{type(self).__name__}({self._end}, {self._free_location}, {self._free_num_bytes}, {self._free_num_slices}, {self._begin_num_bytes}, {self._compression!r}, {self._info_location}, {self._info_num_bytes}, {self._uuid!r})"
 
     @property
     def end(self):
@@ -2252,13 +2210,7 @@ class CascadingFile:
         self._tlist_of_streamers = tlist_of_streamers
 
     def __repr__(self):
-        return "{}({}, {}, {}, {})".format(
-            type(self).__name__,
-            self._fileheader,
-            self._streamers,
-            self._freesegments,
-            self._rootdirectory,
-        )
+        return f"{type(self).__name__}({self._fileheader}, {self._streamers}, {self._freesegments}, {self._rootdirectory})"
 
     @property
     def fileheader(self):
@@ -2292,10 +2244,7 @@ def create_empty(
     Function to create an empty ROOT file, returning a :doc:`uproot.writing._cascade.CascadingFile`.
     """
     filename = sink.file_path
-    if filename is None:
-        filename = "dynamic.root"
-    else:
-        filename = os.path.split(filename)[-1]
+    filename = "dynamic.root" if filename is None else os.path.split(filename)[-1]
     if len(filename) >= 256:
         raise ValueError("ROOT file names must be less than 256 bytes")
 
@@ -2420,9 +2369,7 @@ def update_existing(sink, initial_directory_bytes, uuid_function):
     )
     if raw_bytes[:4] != b"root":
         raise ValueError(
-            "not a ROOT file: first four bytes are {}{}".format(
-                repr(raw_bytes[:4]), sink.in_path
-            )
+            f"not a ROOT file: first four bytes are {raw_bytes[:4]!r}{sink.in_path}"
         )
     fileheader = FileHeader.deserialize(raw_bytes, 0)
 
