@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This is an internal module for writing TTrees in the "cascading" file writer. TTrees
@@ -13,7 +13,7 @@ and sometimes freeing data.
 
 See :doc:`uproot.writing._cascade` for a general overview of the cascading writer concept.
 """
-
+from __future__ import annotations
 
 import datetime
 import math
@@ -40,6 +40,7 @@ _dtype_to_char = {
     numpy.dtype(">u8"): "l",
     numpy.dtype(">f4"): "F",
     numpy.dtype(">f8"): "D",
+    numpy.dtype(">U"): "C",
 }
 
 
@@ -110,7 +111,7 @@ class Tree:
             branch_datashape = None
 
             if isinstance(branch_type, Mapping) and all(
-                uproot._util.isstr(x) for x in branch_type
+                isinstance(x, str) for x in branch_type
             ):
                 branch_dict = branch_type
 
@@ -118,35 +119,34 @@ class Tree:
                 try:
                     if uproot._util.from_module(branch_type, "awkward"):
                         raise TypeError
-                    if (
-                        uproot._util.isstr(branch_type)
-                        and branch_type.strip() == "bytes"
-                    ):
+                    if isinstance(branch_type, str) and branch_type.strip() == "bytes":
                         raise TypeError
                     branch_dtype = numpy.dtype(branch_type)
 
-                except TypeError as err:
+                except (TypeError, ValueError) as err:
                     try:
                         awkward = uproot.extras.awkward()
                     except ModuleNotFoundError as err:
                         raise TypeError(
                             f"not a NumPy dtype and 'awkward' cannot be imported: {branch_type!r}"
                         ) from err
-                    if isinstance(branch_type, awkward.types.Type):
+                    if isinstance(
+                        branch_type,
+                        (awkward.types.Type, awkward.types.ArrayType),
+                    ):
                         branch_datashape = branch_type
                     else:
                         try:
-                            branch_datashape = awkward.types.from_datashape(branch_type)
+                            branch_datashape = awkward.types.from_datashape(
+                                branch_type, highlevel=False
+                            )
                         except Exception:
                             raise TypeError(
                                 f"not a NumPy dtype or an Awkward datashape: {branch_type!r}"
                             ) from err
-                    # checking by class name to be Awkward v1/v2 insensitive
-                    if type(branch_datashape).__name__ == "ArrayType":
-                        if hasattr(branch_datashape, "content"):
-                            branch_datashape = branch_datashape.content
-                        else:
-                            branch_datashape = branch_datashape.type
+                    if isinstance(branch_datashape, awkward.types.ArrayType):
+                        branch_datashape = branch_datashape.content
+
                     branch_dtype = self._branch_ak_to_np(branch_datashape)
 
             if branch_dict is not None:
@@ -166,9 +166,7 @@ class Tree:
                             dtype = numpy.dtype(content)
                         except Exception as err:
                             raise TypeError(
-                                "values of a dict must be NumPy types\n\n    key {} has type {}".format(
-                                    repr(key), repr(content)
-                                )
+                                f"values of a dict must be NumPy types\n\n    key {key!r} has type {content!r}"
                             ) from err
                         self._branch_lookup[subname] = len(self._branch_data)
                         self._branch_data.append(
@@ -188,17 +186,18 @@ class Tree:
                     parameters = {}
 
                 if parameters.get("__array__") == "string":
-                    raise NotImplementedError("array of strings")
+                    if branch_name not in self._branch_lookup:
+                        self._branch_lookup[branch_name] = len(self._branch_data)
+                        self._branch_data.append(
+                            self._branch_np(branch_name, branch_type, numpy.dtype(str))
+                        )
 
                 elif parameters.get("__array__") == "bytes":
                     raise NotImplementedError("array of bytes")
 
-                # checking by class name to be Awkward v1/v2 insensitive
+                # 'awkward' is not in namespace
                 elif type(branch_datashape).__name__ == "ListType":
-                    if hasattr(branch_datashape, "content"):
-                        content = branch_datashape.content
-                    else:
-                        content = branch_datashape.type
+                    content = branch_datashape.content
 
                     counter_name = self._counter_name(branch_name)
                     counter_dtype = numpy.dtype(numpy.int32)
@@ -216,7 +215,7 @@ class Tree:
                             contents = content.contents
                         else:
                             contents = content.fields()
-                        keys = content.keys
+                        keys = content.fields
                         if callable(keys):
                             keys = keys()
                         if keys is None:
@@ -233,9 +232,7 @@ class Tree:
                                 dtype = self._branch_ak_to_np(cont)
                                 if dtype is None:
                                     raise TypeError(
-                                        "fields of a record must be NumPy types, though the record itself may be in a jagged array\n\n    field {} has type {}".format(
-                                            repr(key), str(cont)
-                                        )
+                                        f"fields of a record must be NumPy types, though the record itself may be in a jagged array\n\n    field {key!r} has type {cont!s}"
                                     )
                                 if subname not in self._branch_lookup:
                                     self._branch_lookup[subname] = len(
@@ -251,9 +248,7 @@ class Tree:
                         dt = self._branch_ak_to_np(content)
                         if dt is None:
                             raise TypeError(
-                                "cannot write Awkward Array type to ROOT file:\n\n    {}".format(
-                                    str(branch_datashape)
-                                )
+                                f"cannot write Awkward Array type to ROOT file:\n\n    {branch_datashape!s}"
                             )
                         if branch_name not in self._branch_lookup:
                             self._branch_lookup[branch_name] = len(self._branch_data)
@@ -266,7 +261,7 @@ class Tree:
                         contents = branch_datashape.contents
                     else:
                         contents = branch_datashape.fields()
-                    keys = branch_datashape.keys
+                    keys = branch_datashape.fields
                     if callable(keys):
                         keys = keys()
                     if keys is None:
@@ -283,9 +278,7 @@ class Tree:
                             dtype = self._branch_ak_to_np(content)
                             if dtype is None:
                                 raise TypeError(
-                                    "fields of a record must be NumPy types, though the record itself may be in a jagged array\n\n    field {} has type {}".format(
-                                        repr(key), str(content)
-                                    )
+                                    f"fields of a record must be NumPy types, though the record itself may be in a jagged array\n\n    field {key!r} has type {content!s}"
                                 )
                             if subname not in self._branch_lookup:
                                 self._branch_lookup[subname] = len(self._branch_data)
@@ -295,9 +288,7 @@ class Tree:
 
                 else:
                     raise TypeError(
-                        "cannot write Awkward Array type to ROOT file:\n\n    {}".format(
-                            str(branch_datashape)
-                        )
+                        f"cannot write Awkward Array type to ROOT file:\n\n    {branch_datashape!s}"
                     )
 
         self._num_entries = 0
@@ -325,23 +316,21 @@ class Tree:
         self._key = None
 
     def _branch_ak_to_np(self, branch_datashape):
-        # checking by class name to be Awkward v1/v2 insensitive
-        if type(branch_datashape).__name__ == "NumpyType":
+        if type(branch_datashape).__name__ == "UnknownType":
+            return numpy.dtype("float64")
+        elif type(branch_datashape).__name__ == "NumpyType":
             return numpy.dtype(branch_datashape.primitive)
         elif type(branch_datashape).__name__ == "PrimitiveType":
             return numpy.dtype(branch_datashape.dtype)
         elif type(branch_datashape).__name__ == "RegularType":
-            if hasattr(branch_datashape, "content"):
-                content = self._branch_ak_to_np(branch_datashape.content)
-            else:
-                content = self._branch_ak_to_np(branch_datashape.type)
+            content = self._branch_ak_to_np(branch_datashape.content)
             if content is None:
                 return None
             elif content.subdtype is None:
                 dtype, shape = content, ()
             else:
                 dtype, shape = content.subdtype
-            return numpy.dtype((dtype, (branch_datashape.size,) + shape))
+            return numpy.dtype((dtype, (branch_datashape.size, *shape)))
         else:
             return None
 
@@ -513,7 +502,9 @@ class Tree:
         if uproot._util.from_module(data, "pandas"):
             import pandas
 
-            if isinstance(data, pandas.DataFrame) and data.index.is_numeric():
+            if isinstance(
+                data, pandas.DataFrame
+            ) and uproot._util.pandas_has_attr_is_numeric(pandas)(data.index):
                 provided = dataframe_to_dict(data)
 
         if uproot._util.from_module(data, "awkward"):
@@ -541,7 +532,7 @@ class Tree:
 
         if provided is None:
             if not isinstance(data, Mapping) or not all(
-                uproot._util.isstr(x) for x in data
+                isinstance(x, str) for x in data
             ):
                 raise TypeError(
                     "'extend' requires a mapping from branch name (str) to arrays"
@@ -558,7 +549,7 @@ class Tree:
                                 warnings.simplefilter(
                                     "error", category=numpy.VisibleDeprecationWarning
                                 )
-                                v = numpy.array(v)
+                                v = numpy.array(v)  # noqa: PLW2901 (overwriting v)
                             if v.dtype == numpy.dtype("O"):
                                 raise Exception
                         except (numpy.VisibleDeprecationWarning, Exception):
@@ -568,7 +559,7 @@ class Tree:
                                 raise TypeError(
                                     f"NumPy dtype would be dtype('O'), so we won't use NumPy, but 'awkward' cannot be imported: {k}: {type(v)}"
                                 ) from err
-                            v = awkward.from_iter(v)
+                            v = awkward.from_iter(v)  # noqa: PLW2901 (overwriting v)
 
                     if getattr(v, "dtype", None) == numpy.dtype("O"):
                         try:
@@ -577,7 +568,7 @@ class Tree:
                             raise TypeError(
                                 f"NumPy dtype is dtype('O'), so we won't use NumPy, but 'awkward' cannot be imported: {k}: {type(v)}"
                             ) from err
-                        v = awkward.from_iter(v)
+                        v = awkward.from_iter(v)  # noqa: PLW2901 (overwriting v)
 
                 if uproot._util.from_module(v, "awkward"):
                     try:
@@ -593,22 +584,16 @@ class Tree:
                     ):
                         kk = self._counter_name(k)
                         vv = numpy.asarray(awkward.num(v, axis=1), dtype=">u4")
-                        if kk in provided:
-                            if not numpy.array_equal(vv, provided[kk]):
-                                raise ValueError(
-                                    "branch {} provided both as an explicit array and generated as a counter, and they disagree".format(
-                                        repr(kk)
-                                    )
-                                )
+                        if kk in provided and not numpy.array_equal(vv, provided[kk]):
+                            raise ValueError(
+                                f"branch {kk!r} provided both as an explicit array and generated as a counter, and they disagree"
+                            )
                         provided[kk] = vv
 
-                if k in provided:
-                    if not numpy.array_equal(v, provided[k]):
-                        raise ValueError(
-                            "branch {} provided both as an explicit array and generated as a counter, and they disagree".format(
-                                repr(kk)
-                            )
-                        )
+                if k in provided and not numpy.array_equal(v, provided[k]):
+                    raise ValueError(
+                        f"branch {kk!r} provided both as an explicit array and generated as a counter, and they disagree"
+                    )
                 provided[k] = v
 
         actual_branches = {}
@@ -668,10 +653,7 @@ class Tree:
                 num_entries = len(branch_array)
             elif num_entries != len(branch_array):
                 raise ValueError(
-                    "'extend' must fill every branch with the same number of entries; {} has {} entries".format(
-                        repr(branch_name),
-                        len(branch_array),
-                    )
+                    f"'extend' must fill every branch with the same number of entries; {branch_name!r} has {len(branch_array)} entries"
                 )
 
             datum = self._branch_data[self._branch_lookup[branch_name]]
@@ -679,20 +661,64 @@ class Tree:
                 continue
 
             if datum["counter"] is None:
-                big_endian = numpy.asarray(branch_array, dtype=datum["dtype"])
-                if big_endian.shape != (len(branch_array),) + datum["shape"]:
-                    raise ValueError(
-                        "'extend' must fill branches with a consistent shape: has {}, trying to fill with {}".format(
-                            datum["shape"],
-                            big_endian.shape[1:],
+                if datum["dtype"] == ">U0":
+                    lengths = numpy.asarray(awkward.num(branch_array.layout))
+                    which_big = lengths >= 255
+
+                    lengths_extension_offsets = numpy.empty(
+                        len(branch_array.layout) + 1, numpy.int64
+                    )
+                    lengths_extension_offsets[0] = 0
+                    numpy.cumsum(which_big * 4, out=lengths_extension_offsets[1:])
+
+                    lengths_extension = awkward.contents.ListOffsetArray(
+                        awkward.index.Index64(lengths_extension_offsets),
+                        awkward.contents.NumpyArray(
+                            lengths[which_big].astype(">u4").view("u1")
+                        ),
+                    )
+
+                    lengths[which_big] = 255
+
+                    leafc_data_awkward = awkward.concatenate(
+                        [
+                            lengths.reshape(-1, 1).astype("u1"),
+                            lengths_extension,
+                            awkward.without_parameters(branch_array.layout),
+                        ],
+                        axis=1,
+                    )
+
+                    big_endian = numpy.asarray(awkward.flatten(leafc_data_awkward))
+                    big_endian_offsets = (
+                        lengths_extension_offsets
+                        + numpy.asarray(branch_array.layout.offsets)
+                        + numpy.arange(len(branch_array.layout.offsets))
+                    ).astype(">i4", copy=True)
+                    tofill.append(
+                        (
+                            branch_name,
+                            datum["compression"],
+                            big_endian,
+                            big_endian_offsets,
                         )
                     )
-                tofill.append((branch_name, datum["compression"], big_endian, None))
-
-                if datum["kind"] == "counter":
-                    datum["tleaf_maximum_value"] = max(
-                        big_endian.max(), datum["tleaf_maximum_value"]
+                else:
+                    big_endian = uproot._util.ensure_numpy(branch_array).astype(
+                        datum["dtype"]
                     )
+                    if big_endian.shape != (len(branch_array),) + datum["shape"]:
+                        raise ValueError(
+                            "'extend' must fill branches with a consistent shape: has {}, trying to fill with {}".format(
+                                datum["shape"],
+                                big_endian.shape[1:],
+                            )
+                        )
+                    tofill.append((branch_name, datum["compression"], big_endian, None))
+                    if datum["kind"] == "counter":
+                        datum["tleaf_maximum_value"] = max(
+                            big_endian.max(), datum["tleaf_maximum_value"]
+                        )
 
             else:
                 try:
@@ -702,39 +728,12 @@ class Tree:
                         f"a jagged array was provided (possibly as an iterable), but 'awkward' cannot be imported: {branch_name}: {branch_array!r}"
                     ) from err
                 layout = branch_array.layout
-                while not isinstance(
-                    layout,
-                    (
-                        awkward.layout.ListOffsetArray32,
-                        awkward.layout.ListOffsetArrayU32,
-                        awkward.layout.ListOffsetArray64,
-                    ),
-                ):
-                    if isinstance(layout, awkward.partition.PartitionedArray):
-                        layout = awkward.concatenate(layout.partitions, highlevel=False)
-
-                    elif isinstance(
-                        layout,
-                        (
-                            awkward.layout.IndexedArray32,
-                            awkward.layout.IndexedArrayU32,
-                            awkward.layout.IndexedArray64,
-                        ),
-                    ):
+                while not isinstance(layout, awkward.contents.ListOffsetArray):
+                    if isinstance(layout, awkward.contents.IndexedArray):
                         layout = layout.project()
 
-                    elif isinstance(
-                        layout,
-                        (
-                            awkward.layout.ListArray32,
-                            awkward.layout.ListArrayU32,
-                            awkward.layout.ListArray64,
-                        ),
-                    ):
-                        layout = layout.toListOffsetArray64(False)
-
-                    elif isinstance(layout, awkward.layout.VirtualArray):
-                        layout = layout.array
+                    elif isinstance(layout, awkward.contents.ListArray):
+                        layout = layout.to_ListOffsetArray64(False)
 
                     else:
                         raise AssertionError(
@@ -751,33 +750,23 @@ class Tree:
                     content = content[: offsets[-1]]
 
                 shape = [len(content)]
-                while not isinstance(content, awkward.layout.NumpyArray):
-                    if isinstance(
-                        content,
-                        (
-                            awkward.layout.IndexedArray32,
-                            awkward.layout.IndexedArrayU32,
-                            awkward.layout.IndexedArray64,
-                        ),
-                    ):
+                while not isinstance(content, awkward.contents.NumpyArray):
+                    if isinstance(content, awkward.contents.IndexedArray):
                         content = content.project()
 
-                    elif isinstance(content, awkward.layout.EmptyArray):
-                        content = content.toNumpyArray()
+                    elif isinstance(content, awkward.contents.EmptyArray):
+                        content = content.to_NumpyArray(dtype=numpy.float64)
 
-                    elif isinstance(content, awkward.layout.RegularArray):
+                    elif isinstance(content, awkward.contents.RegularArray):
                         shape.append(content.size)
                         content = content.content
 
-                    elif isinstance(layout, awkward.layout.VirtualArray):
-                        content = content.array
-
                     else:
                         raise AssertionError(
                             "how did this pass the type check?\n\n" + repr(content)
                         )
 
-                big_endian = numpy.asarray(content, dtype=datum["dtype"])
+                big_endian = numpy.asarray(content.data, dtype=datum["dtype"])
                 shape = tuple(shape) + big_endian.shape[1:]
 
                 if shape[1:] != datum["shape"]:
@@ -804,7 +793,13 @@ class Tree:
         for branch_name, compression, big_endian, big_endian_offsets in tofill:
             datum = self._branch_data[self._branch_lookup[branch_name]]
 
-            if big_endian_offsets is None:
+            if datum["dtype"] == ">U0":
+                totbytes, zipbytes, location = self.write_string_basket(
+                    sink, branch_name, compression, big_endian, big_endian_offsets
+                )
+                datum["fEntryOffsetLen"] = 4 * (len(big_endian_offsets) - 1)
+
+            elif big_endian_offsets is None:
                 totbytes, zipbytes, location = self.write_np_basket(
                     sink, branch_name, compression, big_endian
                 )
@@ -859,8 +854,8 @@ class Tree:
         # TAttMarker v2, fMarkerColor: 1, fMarkerStyle: 1, fMarkerSize: 1.0
         out.append(
             b"@\x00\x00\x08\x00\x02\x02Z\x00\x01\x00\x01"
-            + b"@\x00\x00\x06\x00\x02\x00\x00\x03\xe9"
-            + b"@\x00\x00\n\x00\x02\x00\x01\x00\x01?\x80\x00\x00"
+            b"@\x00\x00\x06\x00\x02\x00\x00\x03\xe9"
+            b"@\x00\x00\n\x00\x02\x00\x01\x00\x01?\x80\x00\x00"
         )
 
         metadata_out_index = len(out)
@@ -1001,12 +996,17 @@ class Tree:
                 special_struct = uproot.models.TLeaf._tleafs1_format1
             elif letter_upper == "I":
                 special_struct = uproot.models.TLeaf._tleafi1_format1
+            elif letter_upper == "G":
+                special_struct = uproot.models.TLeaf._tleafl1_format0
             elif letter_upper == "L":
                 special_struct = uproot.models.TLeaf._tleafl1_format0
             elif letter_upper == "F":
                 special_struct = uproot.models.TLeaf._tleaff1_format1
             elif letter_upper == "D":
                 special_struct = uproot.models.TLeaf._tleafd1_format1
+            elif letter_upper == "C":
+                special_struct = uproot.models.TLeaf._tleafc1_format1
+
             fLenType = datum["dtype"].itemsize
             fIsUnsigned = letter != letter_upper
 
@@ -1025,8 +1025,36 @@ class Tree:
             leaf_title_length = (1 if len(leaf_title) < 255 else 5) + len(leaf_title)
 
             leaf_header = numpy.array(
-                [64, 0, 0, 76, 0, 1, 64, 0, 0, 54, 0, 2, 64, 0]
-                + [0, 30, 0, 1, 0, 1, 0, 0, 0, 0, 3, 0, 0, 0],
+                [
+                    64,
+                    0,
+                    0,
+                    76,
+                    0,
+                    1,
+                    64,
+                    0,
+                    0,
+                    54,
+                    0,
+                    2,
+                    64,
+                    0,
+                    0,
+                    30,
+                    0,
+                    1,
+                    0,
+                    1,
+                    0,
+                    0,
+                    0,
+                    0,
+                    3,
+                    0,
+                    0,
+                    0,
+                ],
                 numpy.uint8,
             )
             tmp = leaf_header[0:4].view(">u4")
@@ -1099,12 +1127,12 @@ class Tree:
             out.append(special_struct.pack(0, 0))
             datum["tleaf_special_struct"] = special_struct
 
-            out[
-                subany_tleaf_index
-            ] = uproot.serialization._serialize_object_any_format1.pack(
-                numpy.uint32(sum(len(x) for x in out[subany_tleaf_index + 1 :]) + 4)
-                | uproot.const.kByteCountMask,
-                uproot.const.kNewClassTag,
+            out[subany_tleaf_index] = (
+                uproot.serialization._serialize_object_any_format1.pack(
+                    numpy.uint32(sum(len(x) for x in out[subany_tleaf_index + 1 :]) + 4)
+                    | uproot.const.kByteCountMask,
+                    uproot.const.kNewClassTag,
+                )
             )
 
             out[subtobjarray_of_leaves_index] = uproot.serialization.numbytes_version(
@@ -1141,12 +1169,12 @@ class Tree:
                 sum(len(x) for x in out[tbranch_index + 1 :]), 13  # TBranch
             )
 
-            out[
-                any_tbranch_index
-            ] = uproot.serialization._serialize_object_any_format1.pack(
-                numpy.uint32(sum(len(x) for x in out[any_tbranch_index + 1 :]) + 4)
-                | uproot.const.kByteCountMask,
-                uproot.const.kNewClassTag,
+            out[any_tbranch_index] = (
+                uproot.serialization._serialize_object_any_format1.pack(
+                    numpy.uint32(sum(len(x) for x in out[any_tbranch_index + 1 :]) + 4)
+                    | uproot.const.kByteCountMask,
+                    uproot.const.kNewClassTag,
+                )
             )
 
         out[tobjarray_of_branches_index] = uproot.serialization.numbytes_version(
@@ -1287,6 +1315,27 @@ class Tree:
 
             datum["arrays_write_start"] = datum["arrays_write_stop"]
 
+            if datum["dtype"] == ">U0":
+                position = (
+                    base
+                    + datum["basket_metadata_start"]
+                    - 25  # empty TObjArray of fBaskets (embedded)
+                    - 8  # specialized TLeaf* members (fMinimum, fMaximum)
+                    - 4  # null fLeafCount
+                    - 14  # generic TLeaf members
+                )
+                sink.write(
+                    position,
+                    uproot.models.TLeaf._tleaf2_format0.pack(
+                        self._metadata["fLen"],
+                        datum["dtype"].itemsize,
+                        0,
+                        datum["kind"] == "counter",
+                        _dtype_to_char[datum["dtype"]]
+                        != _dtype_to_char[datum["dtype"]].upper(),
+                    ),
+                )
+
             if datum["kind"] == "counter":
                 position = (
                     base
@@ -1443,6 +1492,96 @@ class Tree:
 
         return fKeylen + fObjlen, fNbytes, location
 
+    def write_string_basket(self, sink, branch_name, compression, array, offsets):
+        fClassName = uproot.serialization.string("TBasket")
+        fName = uproot.serialization.string(branch_name)
+        fTitle = uproot.serialization.string(self._name)
+        fKeylen = (
+            uproot.reading._key_format_big.size
+            + len(fClassName)
+            + len(fName)
+            + len(fTitle)
+            + uproot.models.TBasket._tbasket_format2.size
+            + 1
+        )
+
+        itemsize = array.dtype.itemsize
+        for item in array.shape[1:]:
+            itemsize *= item
+        try:
+            uproot.extras.awkward()
+        except ModuleNotFoundError as err:
+            raise TypeError(
+                f"'awkward' cannot be imported: {self._branch_type!r}"
+            ) from err
+
+        offsets *= itemsize
+        offsets += fKeylen
+
+        raw_array = uproot._util.tobytes(array)
+        raw_offsets = uproot._util.tobytes(offsets)
+        uncompressed_data = (
+            raw_array
+            + _tbasket_offsets_length.pack(len(offsets))
+            + raw_offsets[:-4]
+            + b"\x00\x00\x00\x00"
+        )
+        compressed_data = uproot.compression.compress(uncompressed_data, compression)
+
+        # get size of biggest string
+        self._metadata["fLen"] = (
+            0
+            if len(offsets) == 1
+            else max([offsets[i + 1] - offsets[i] for i in range(len(offsets) - 1)])
+        )
+
+        fLast = offsets[-1]
+        offsets[-1] = 0
+
+        fObjlen = len(uncompressed_data)
+        fNbytes = fKeylen + len(compressed_data)
+
+        parent_location = self._directory.key.location  # FIXME: is this correct?
+
+        location = self._freesegments.allocate(fNbytes, dry_run=False)
+
+        out = []
+        out.append(
+            uproot.reading._key_format_big.pack(
+                fNbytes,
+                1004,  # fVersion
+                fObjlen,
+                uproot._util.datetime_to_code(datetime.datetime.now()),  # fDatime
+                fKeylen,
+                0,  # fCycle
+                location,  # fSeekKey
+                parent_location,  # fSeekPdir
+            )
+        )
+        out.append(fClassName)
+        out.append(fName)
+        out.append(fTitle)
+        out.append(
+            uproot.models.TBasket._tbasket_format2.pack(
+                3,  # fVersion
+                32000,  # fBufferSize
+                1000,  # fNevBufSize
+                len(offsets) - 1,  # fNevBuf
+                fLast,
+            )
+        )
+
+        out.append(b"\x00")  # part of the Key (included in fKeylen, at least)
+
+        out.append(compressed_data)
+
+        sink.write(location, b"".join(out))
+        self._freesegments.write(sink)
+        sink.set_file_length(self._freesegments.fileheader.end)
+        sink.flush()
+
+        return fKeylen + fObjlen, fNbytes, location
+
 
 _tbasket_offsets_length = struct.Struct(">I")
 
